import os
from os.path import abspath
from pathlib import Path
from typing import Literal, Optional, Union
from uuid import uuid4
from warnings import warn

import dask.dataframe as dd
import pandas as pd
from parse import *

from insitupy import __version__
from insitupy._constants import ISPY_METADATA_FILE
from insitupy._core._xenium import (_read_binned_expression,
                                    _read_boundaries_from_xenium,
                                    _read_matrix_from_xenium,
                                    _restructure_transcripts_dataframe)
from insitupy._core.insitudata import InSituData
from insitupy._exceptions import InvalidXeniumDirectory
from insitupy.io.files import read_json
from insitupy.utils.utils import convert_to_list

from .dataclasses import AnnotationsData, CellData, ImageData, RegionsData


def read_xenium(
    path: Union[str, os.PathLike, Path],
    # names: Union[Literal["all", "nuclei"], str] = "all", # here a specific image can be chosen
    nuclei_type: Literal["focus", "mip", ""] = "mip",
    load_cell_segmentation_images: bool = True,
    verbose: bool = True,
    transcript_mode: Literal["pandas", "dask"] = "dask",
    restructure_transcripts: bool = False
    ) -> InSituData:

    path = Path(path) # make sure the path is a pathlib path
    metadata_filename: str = "experiment.xenium"

    if not (path / metadata_filename).exists():
        raise InvalidXeniumDirectory(directory=path)

    # INITIALIZE INSITUDATA
    # initialize the metadata dict
    metadata = {}
    metadata["data"] = {}
    metadata["history"] = {}
    metadata["history"]["cells"] = []
    metadata["history"]["annotations"] = []
    metadata["history"]["regions"] = []

    # check if path exists
    if not path.is_dir():
        raise FileNotFoundError(f"No such directory found: {str(path)}")

    # save paths of this project in metadata
    metadata["path"] = abspath(path).replace("\\", "/")
    metadata["metadata_file"] = metadata_filename

    # read metadata
    metadata["method_params"] = read_json(path / metadata_filename)

    # get slide id and sample id from metadata
    slide_id = metadata["method_params"]["slide_id"]
    sample_id = metadata["method_params"]["region_name"]

    # initialize the uid section
    metadata["uids"] = [str(uuid4())]

    # add method
    metadata["method"] = "Xenium"

    data = InSituData(path=path,
                        metadata=metadata,
                        slide_id=slide_id,
                        sample_id=sample_id,
                        from_insitudata=False,

                        )

    # LOAD CELLS
    if verbose:
        print("Loading cells...", flush=True)
    pixel_size = data.metadata["method_params"]["pixel_size"]
    # read celldata
    matrix = _read_matrix_from_xenium(path=data.path)
    boundaries = _read_boundaries_from_xenium(path=data.path, pixel_size=pixel_size)
    data.cells = CellData(matrix=matrix, boundaries=boundaries)


    # LOAD IMAGES
    if verbose:
        print("Loading images...", flush=True)
    nuclei_file_key = f"morphology_{nuclei_type}_filepath"

    # In v2.0 the "mip" image was removed due to better focusing of the machine.
    # For <v2.0 the function still tries to retrieve the "mip" image but in case this is not found
    # it will retrieve the "focus" image
    if nuclei_type == "mip" and nuclei_file_key not in data.metadata["method_params"]["images"].keys():

        nuclei_type = "focus"
        nuclei_file_key = f"morphology_{nuclei_type}_filepath"

    # if names == "nuclei":
    img_keys = [nuclei_file_key]
    img_names = ["nuclei"]

    # get path of image files
    img_files = [data.metadata["method_params"]["images"][k] for k in img_keys]

    if load_cell_segmentation_images:
        # get cell segmentation images if available
        if "morphology_focus/" in data.metadata["method_params"]["images"][nuclei_file_key]:
            seg_files = ["morphology_focus/morphology_focus_0001.ome.tif",
                            "morphology_focus/morphology_focus_0002.ome.tif",
                            "morphology_focus/morphology_focus_0003.ome.tif"
                            ]
            seg_names = ["cellseg1", "cellseg2", "cellseg3"]

            # check which segmentation files exist and append to image list
            seg_file_exists_list = [(data.path / f).is_file() for f in seg_files]
            #print(seg_file_exists_list)
            img_files += [f for f, exists in zip(seg_files, seg_file_exists_list) if exists]
            img_names += [n for n, exists in zip(seg_names, seg_file_exists_list) if exists]

    # create imageData object
    img_paths = [data.path / elem for elem in img_files]

    if data.images is None:
        data.images = ImageData(img_paths, img_names)
    else:
        for im, n in zip(img_paths, img_names):
            data.images.add_image(im, n, overwrite=False, verbose=True)

    # LOAD TRANSCRIPTS
    transcript_filename = "transcripts.parquet"
    if verbose:
        print("Loading transcripts...", flush=True)

    if transcript_mode == "pandas":
        transcript_dataframe = pd.read_parquet(data.path / transcript_filename)

        if restructure_transcripts:
            data.transcripts = _restructure_transcripts_dataframe(transcript_dataframe)
        else:
            data.transcripts = transcript_dataframe
    elif transcript_mode == "dask":
        # Load the transcript data using Dask
        data.transcripts = dd.read_parquet(data.path / transcript_filename)
    else:
        raise ValueError(f"Invalid value for `transcript_mode`: {transcript_mode}")

    return data



import json
import os
from numbers import Number
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

import numpy as np
import pandas as pd
import requests
from anndata import AnnData
from matplotlib import colormaps
from scipy.cluster.hierarchy import (dendrogram, fcluster, linkage,
                                     set_link_color_palette)
from scipy.spatial.distance import squareform

#from .adata import create_deg_df

class SpeciesToID:
    def __init__(self):
        self.species_dict = {
            'mmusculus': 10090,
            'hsapiens': 9606,
            'dmelanogaster': 7227
        }
    def check_species(self, species):
        if species not in self.species_dict:
            raise ValueError(
                "`species` must be one of following values: {}".format(list(self.species_dict.keys()))
            )
    def convert(self, species):
        self.check_species(species)
        return self.species_dict[species]

def find_between(s, first, last ):
    try:
        start = s.index( first ) + len( first )
        end = s.index( last, start )
        return s[start:end]
    except ValueError:
        return ""

class GOEnrichment():
    def __init__(self):
        self._results = {}

    @property
    def results(self):
        return self._results

    def gprofiler(self, target_genes: Union[dict, list] = None, top_n: Optional[int] = None,
                  organism: str = None, key_added: str = 'result',
                  uns_key_added: str = 'gprofiler', return_df: bool = True,
                  sortby: str = 'pvals_adj', **kwargs: Any):
        """
        Performs GO term enrichment analysis using the gprofiler web resource.

        Args:
            target_genes (Union[dict, list]): A dictionary where keys are query names and values are lists of genes, or a list of genes.
            top_n (int, optional): The number of top genes to consider for enrichment analysis. Defaults to 300.
            organism (str): The organism name following gprofiler naming conventions (e.g., 'mmusculus').
            key_added (str, optional): The key under which to save the results. Defaults to None.
            uns_key_added (str, optional): The key under which to save the results in the results dictionary. Defaults to 'gprofiler'.
            return_df (bool, optional): Whether to return the results as a DataFrame. Defaults to True.
            sortby (str, optional): The column to sort the results by. Defaults to 'pvals_adj'.
            **kwargs (Any): Additional arguments to pass to the gprofiler function.

        Returns:
            pd.DataFrame: A DataFrame containing the enrichment results if return_df is True.

        Raises:
            ValueError: If `target_genes` is not a dictionary or list, or if `organism` is not specified.
        """
        try:
            from gprofiler import GProfiler
        except ImportError:
            raise ImportError("This function requires the 'gprofiler' package. Please install it with 'pip install gprofiler-official'.")

        if isinstance(target_genes, dict):
            groups = target_genes.keys()
        elif isinstance(target_genes, list):
            groups = ['query']
            target_genes = {'query': target_genes}
        else:
            raise ValueError("`target_genes` must be a dictionary with query names as keys and lists of genes as values, or a list of genes.")

        if organism is None:
            raise ValueError("`organism` not specified. Needs gprofiler naming conventions, e.g. `mmusculus`")

        enrichment_dict = {}
        for group, genes in target_genes.items():
            if top_n is not None:
                genes = genes[:top_n]

            gp = GProfiler(return_dataframe=True)
            e = gp.profile(organism=organism, query=genes, no_evidences=False)

            # calc -log(p_value)
            e['Enrichment score'] = [-np.log10(elem) for elem in e.p_value]

            # sort by p_value
            e.sort_values('p_value', inplace=True)

            # collect data
            enrichment_dict[group] = e

        enrichment = pd.concat(enrichment_dict)
        # rename column headers
        enrichment.rename(columns={'recall': 'Gene ratio'}, inplace=True)

        # save in class
        if not uns_key_added in self._results:
            self._results[uns_key_added] = {}

        self._results[uns_key_added][key_added] = enrichment

        if return_df:
            return enrichment

    def stringdb(self, target_genes: Union[dict, list] = None, top_n: Optional[int] = None,
                 organism: str = None, key_added: str = 'result',
                 uns_key_added: str = 'stringdb', return_df: bool = True,
                 sortby: str = 'pvals_adj', **kwargs: Any):
        """
        Performs GO term enrichment analysis using the stringdb web resource.

        Args:
            target_genes (Union[dict, list]): A dictionary where keys are query names and values are lists of genes, or a list of genes.
            top_n (int, optional): The number of top genes to consider for enrichment analysis. Defaults to 300.
            organism (str): The organism name following stringdb naming conventions (e.g., 'mmusculus').
            key_added (str, optional): The key under which to save the results. Defaults to None.
            uns_key_added (str, optional): The key under which to save the results in the results dictionary. Defaults to 'stringdb'.
            return_df (bool, optional): Whether to return the results as a DataFrame. Defaults to True.
            sortby (str, optional): The column to sort the results by. Defaults to 'pvals_adj'.
            **kwargs (Any): Additional arguments to pass to the stringdb function.

        Returns:
            pd.DataFrame: A DataFrame containing the enrichment results if return_df is True.

        Raises:
            ValueError: If `target_genes` is not a dictionary or list, or if `organism` is not specified.
        """
        if isinstance(target_genes, dict):
            groups = target_genes.keys()
        elif isinstance(target_genes, list):
            groups = ['query']
            target_genes = {'query': target_genes}
        else:
            raise ValueError("`target_genes` must be a dictionary with query names as keys and lists of genes as values, or a list of genes.")

        if organism is None:
            raise ValueError("`organism` not specified. Needs stringdb naming conventions, e.g. `mmusculus`")

        enrichment_dict = {}
        for group, genes in target_genes.items():
            if top_n is not None:
                genes = genes[:top_n]

            sdb = StringDB()
            sdb.call_stringdb_enrichment(genes=genes, species=organism)
            e = sdb.result

            # modify data to fit into further analysis
            e.rename(columns={
                "category": "source",
                "description": "name",
                "term": "native",
                "inputGenes": "intersections"
            }, inplace=True)

            # sort by Enrichment score if it exists
            if 'Enrichment score' in e.columns:
                e.sort_values('Enrichment score', inplace=True, ascending=False)

            # collect data
            enrichment_dict[group] = e

        enrichment = pd.concat(enrichment_dict)
        # rename column headers
        enrichment.rename(columns={'recall': 'Gene ratio'}, inplace=True)

        # save in class
        if not uns_key_added in self._results:
            self._results[uns_key_added] = {}

        self._results[uns_key_added][key_added] = enrichment

        if return_df:
            return enrichment

    def enrichr(self, target_genes: Union[dict, list] = None, top_n: Optional[int] = None,
                organism: str = None, key_added: str = 'result',
                enrichr_libraries: str = 'GO_Biological_Process_2018',
                outdir: str = None, no_plot: bool = True,
                uns_key_added: str = 'enrichr', return_df: bool = True,
                sortby: str = 'pvals_adj', **kwargs: Any):
        """
        Performs GO term enrichment analysis using the enrichr web resource.

        Args:
            target_genes (Union[dict, list]): A dictionary where keys are query names and values are lists of genes, or a list of genes.
            top_n (int, optional): The number of top genes to consider for enrichment analysis. Defaults to 300.
            organism (str): The organism name following enrichr naming conventions (e.g., 'mouse', 'human').
            key_added (str, optional): The key under which to save the results. Defaults to None.
            enrichr_libraries (str, optional): The Enrichr libraries to use for the analysis. Defaults to 'GO_Biological_Process_2018'.
            outdir (str, optional): The output directory for Enrichr results. Defaults to None.
            no_plot (bool, optional): Whether to suppress plot generation. Defaults to True.
            uns_key_added (str, optional): The key under which to save the results in the results dictionary. Defaults to 'enrichr'.
            return_df (bool, optional): Whether to return the results as a DataFrame. Defaults to True.
            sortby (str, optional): The column to sort the results by. Defaults to 'pvals_adj'.
            **kwargs (Any): Additional arguments to pass to the enrichr function.

        Returns:
            pd.DataFrame: A DataFrame containing the enrichment results if return_df is True.

        Raises:
            ValueError: If `target_genes` is not a dictionary or list, or if `organism` is not specified.
        """
        try:
            import gseapy
        except ImportError:
            raise ImportError("This function requires the 'gseapy' package. Please install it with 'pip install gseapy'.")




        if isinstance(target_genes, dict):
            groups = target_genes.keys()
        elif isinstance(target_genes, list):
            groups = ['query']
            target_genes = {'query': target_genes}
        else:
            raise ValueError("`target_genes` must be a dictionary with query names as keys and lists of genes as values, or a list of genes.")

        if organism is None:
            raise ValueError("`organism` not specified. Needs to have one of the following values: `mouse`, `human`")

        enrichment_dict = {}
        for group, genes in target_genes.items():
            if top_n is not None:
                genes = genes[:top_n]

            e = gseapy.enrichr(gene_list=genes, gene_sets=enrichr_libraries,
                            organism=organism, outdir=outdir, no_plot=no_plot,
                            **kwargs).results

            # calc -log(p_value)
            e['Enrichment score'] = [-np.log10(elem) for elem in e['Adjusted P-value']]

            # sort by Enrichment score
            e.sort_values('Enrichment score', inplace=True, ascending=False)

            # calculate gene ratio
            e['Gene ratio'] = [int(elem.split("/")[0]) / int(elem.split("/")[1]) for elem in e['Overlap']]

            # collect data
            enrichment_dict[group] = e

        enrichment = pd.concat(enrichment_dict)

        # rename column headers
        enrichment.rename(columns={
            'Term': 'name',
            'Gene_set': 'source',
            'Genes': 'intersections'
        }, inplace=True)

        # transform intersections into list
        enrichment['intersections'] = [elem.split(";") for elem in enrichment['intersections']]

        # separate human-readable name from GO ID
        enrichment['native'] = [elem.split(" (")[1].rstrip(")") if " (GO" in elem else np.nan for elem in enrichment['name']]
        enrichment['name'] = [elem.split(" (")[0] if " (GO" in elem else elem for elem in enrichment['name']]

        # save in class
        if not uns_key_added in self._results:
            self._results[uns_key_added] = {}

        self._results[uns_key_added][key_added] = enrichment

        if return_df:
            return enrichment


    def __repr__(self):
        repr_str = "GOEnrichment analyses performed:\n"
        for uns_key, analyses in self._results.items():
            repr_str += f"  {uns_key}:\n"
            for key in analyses.keys():
                repr_str += f"    - {key}\n"
        return repr_str


class StringDB:
    def __init__(self, return_results: bool = True):
        self.result = None
        self.return_results = return_results

    def call_stringdb_enrichment(self, genes, species):
        '''
        Function to get functional enrichment results from https://string-db.org/
        Code based on https://string-db.org/help/api/
        '''

        ## Settings to call string-db
        string_api_url = "https://version-11-5.string-db.org/api"
        output_format = "json"
        method = "enrichment"
        tax_id = SpeciesToID().convert(species)

        ## Construct the request
        request_url = "/".join([string_api_url, output_format, method])

        while True:
            ## Set parameters
            params = {

                "identifiers" : "%0d".join(genes), # your protein
                "species" : tax_id, # species NCBI identifier
                "caller_identity" : "www.awesome_app.org" # your app name
            }

            ## Call STRING
            response = requests.post(request_url, data=params)
            response = json.loads(response.text)

            # make sure STRING found all genes
            try:
                ## Read and parse the results
                self.result = pd.DataFrame(response)
            except ValueError:
                if response['Error'] == 'not found':
                    # extract error message and identify missing gene that caused the error
                    ermsg = response['ErrorMessage']
                    missing_gene = find_between(ermsg, first="called '", last="' in the")

                    # remove missing gene from list
                    genes.remove(missing_gene)
                    print("Gene '{}' was not found by STRING and was removed from query.".format(missing_gene))
            else:
                break

        # rename columns to align them to gprofiler results
        self.result.rename(columns={
            "category": "source",
            "description": "name",
            "term": "native"
            }, inplace=True)

        # calculate enrichment score

        if "fdr" in self.result.columns:
            self.result['Enrichment score'] = [-np.log10(elem) for elem in self.result["fdr"]]
            self.result["Gene ratio"] = [a/b for a,b in zip(self.result["number_of_genes"], self.result["number_of_genes_in_background"])]

        if self.return_results:
            return self.result


    def call_stringdb_network(self, genes, species, output_format="image", prefix="",
        output_dir="stringdb_networks", network_flavor="confidence", save=True, **kwargs):
        '''
        Generate and save networks from https://string-db.org/.
        '''

        # check output format
        format_to_ext = {
            "image": ".png",
            "svg": ".svg",
            }

        if output_format in format_to_ext:
            output_ext = format_to_ext[output_format]
        else:
            raise ValueError("`output_format` must be one of following values: {}".format(list(format_to_ext.keys())))

        string_api_url = "https://version-11-5.string-db.org/api"
        output_format = output_format
        method = "network"
        tax_id = SpeciesToID().convert(species)

        ## Construct URL
        request_url = "/".join([string_api_url, output_format, method])

        ## Set parameters
        params = {

            "identifiers" : "%0d".join(genes), # your protein
            "species" : tax_id, # species NCBI identifier
            #"add_white_nodes": 15, # add 15 white nodes to my protein
            "network_flavor": network_flavor, # show confidence links
            "caller_identity" : "www.awesome_app.org" # your app name
        }

        ## Call STRING
        response = requests.post(request_url, data=params)
        self.result = response.content

        if save:
            ## Save the network to file
            # create output directory
            Path(output_dir).mkdir(parents=True, exist_ok=True)

            output_file = os.path.join(output_dir, "{}_network{}".format(prefix, output_ext))
            print("Saving interaction network to {}".format(output_file))

            with open(output_file, 'wb') as fh:
                fh.write(self.result)

        if self.return_results:
            return self.result

    def stringdb_network_from_adata(self, adata: AnnData = None, key: str = None, top_n: Optional[int] = None, organism: str = None, output_format: str = "image",
        key_added: str = None, sortby: str = 'pvals_adj', ascending: bool = True,
        **kwargs: Any):

        deg, groups, key_added = GOEnrichment().prepare_enrichment(adata=adata, key=key, key_added=key_added,
                        sortby=sortby, ascending=ascending)

        for i, group in enumerate(groups):
            #target_genes = deg.xs((group, 'up'), level=(0,1)).names.tolist()
            if key is not None:
                target_genes = deg.xs(group).names.tolist()

            if top_n is not None:
                target_genes = target_genes[:top_n]

            prefix = "KEY-{}-GROUP-{}".format(key, group)

            sdb = StringDB(return_results=False)
            sdb.call_stringdb_network(genes=target_genes, species=organism, prefix=prefix, output_format=output_format, save=True, **kwargs)



def get_up_down_genes(dge_results,
                      pval_threshold: Number = 0.05,
                      logfold_threshold: Number = 1
                      ):
    pval_mask = dge_results['pvals'] < pval_threshold
    lfc_mask_up = dge_results['logfoldchanges'] > logfold_threshold
    lfc_mask_down = dge_results['logfoldchanges'] < -logfold_threshold

    genes_up = dge_results[lfc_mask_up & pval_mask]['gene'].tolist()
    genes_down = dge_results[lfc_mask_down & pval_mask]['gene'].tolist()

    return genes_up, genes_down


import os
import warnings
import zipfile
from contextlib import ExitStack
from pathlib import Path
from typing import List, Literal, Optional, Union

import dask.array as da
import numpy as np
import xmltodict
import zarr
from parse import *
from tifffile import TiffFile, TiffWriter, imread

from insitupy import __version__

from .._exceptions import InvalidFileTypeError
from ..images.utils import create_img_pyramid
from ..utils.utils import convert_to_list


def read_zarr(path):
    # load image from .zarr.zip
    #zipped = True if suffix == "zarr.zip" else False
    zipped = zipfile.is_zipfile(path)
    with zarr.ZipStore(path, mode="r") if zipped else zarr.DirectoryStore(path) as dirstore:

        # get components of zip store
        components = dirstore.listdir()

        if ".zarray" in components:
            # the store is an array which can be opened
            if zipped:
                img = da.from_zarr(dirstore).persist()
            else:
                img = da.from_zarr(dirstore)
        else:
            subres = [elem for elem in components if not elem.startswith(".")]
            img = []
            for s in subres:
                if zipped:
                    img.append(
                        da.from_zarr(dirstore, component=s).persist()
                                )
                else:
                    img.append(
                        da.from_zarr(dirstore, component=s)
                                )

        # retrieve OME metadata
        store = zarr.open(dirstore)
        meta = store.attrs.asdict()
        ome_meta = meta["OME"]
        axes = meta["axes"]

    return img, ome_meta, axes


def read_image(
    path
    ):
    path = Path(path)
    suffix = path.name.split(".", maxsplit=1)[-1]

    if "zarr" in suffix:
        img, ome_meta, axes = read_zarr(path)

    elif suffix in ["ome.tif", "ome.tiff"]:
        # load image from .ome.tiff
        img = read_ome_tiff(path=path, levels=None)
        # read ome metadata
        with TiffFile(path) as tif:
            axes = tif.series[0].axes # get axes
            ome_meta = tif.ome_metadata # read OME metadata
            ome_meta = xmltodict.parse(ome_meta, attr_prefix="")["OME"] # convert XML to dict

    else:
        raise InvalidFileTypeError(
            allowed_types=["zarr", "zarr.zip", "ome.tif", "ome.tiff"],
            received_type=suffix
            )

    return img, ome_meta, axes

def write_zarr(image, file,
               img_metadata: dict,
               save_pyramid: bool = True,
               axes: str = "YXS", # channels - other examples: 'TCYXS'. S for RGB channels. 'YX' for grayscale image.
               overwrite: bool = False,
               verbose: bool = False
               ):
    if verbose:
        print(f"Saving image to {str(file)}")

    # get suffix
    file = Path(file)

    if file.exists():
        if overwrite:
            file.unlink() # delete file
        else:
            raise FileExistsError("Output file exists already ({}).\nFor overwriting it, select `overwrite=True`".format(file))

    suffix = file.name.split(".", 1)[-1]

    # check if the suffix contains zip
    zipped = "zip" in suffix

    # decide whether to save as pyramid or not
    if isinstance(image, list):
        if not save_pyramid:
            image_data = image[0]
        else:
            image_data = image
    else:
        if save_pyramid:
            # create img pyramid
            image_data = create_img_pyramid(img=image, nsubres=6, axes=axes)
        else:
            image_data = image

    with zarr.ZipStore(file, mode="w") if zipped else zarr.DirectoryStore(file) as dirstore:
        # check whether to save the image as pyramid or not
        if save_pyramid:
            for i, im in enumerate(image_data):
                im.to_zarr(dirstore, component=str(i))
        else:
            # save image data in zipstore without pyramid
            image_data.to_zarr(dirstore)

        # open zarr store save metadata in zarr store
        store = zarr.open(dirstore, mode="a")
        store.attrs.put(img_metadata)
        # for k,v in img_metadata.items():
        #     store.attrs[k] = v

def write_ome_tiff(
    image: Union[np.ndarray, da.core.Array, List[da.core.Array]],
    file: Union[str, os.PathLike, Path],
    axes: str = "YXS", # channels - other examples: 'TCYXS'. S for RGB channels. 'YX' for grayscale image.
    metadata: dict = {},
    subresolutions = 6,
    subres_steps: int = 2,
    pixelsize: Optional[float] = 1, # defaults to Xenium settings.
    pixelunit: Optional[str] = None, # usually µm
    #significant_bits: Optional[int] = 16,
    photometric: Literal['rgb', 'minisblack', 'maxisblack'] = 'rgb', # before I had rgb here. Xenium doc says minisblack
    tile: tuple = (1024, 1024), # 1024 pixel is optimal for Xenium Explorer
    compression: Literal['jpeg', 'LZW', 'jpeg2000', "ZLIB", None] = 'ZLIB', # jpeg2000 or ZLIB are recommended in the Xenium documentation - ZLIB is faster
    overwrite: bool = False,
    verbose: bool = False
    ):

    '''
    Function to write (pyramidal) OME-TIFF files.
    Code adapted from: https://github.com/cgohlke/tifffile and Xenium docs (see below).

    For parameters optimal for Xenium see: https://www.10xgenomics.com/support/software/xenium-explorer/tutorials/xe-image-file-conversion
    '''
    if verbose:
        print(f"Saving image to {str(file)}")
    # check if the image is an image pyramid
    if isinstance(image, list):
        # if it is a pyramid, select only the highest resolution image
        first_image = image[0]
        image_pyramid = image
    elif isinstance(image, np.ndarray) or isinstance(image, da.core.Array):
        first_image = image
        image_pyramid = create_img_pyramid(img=image, nsubres=subresolutions, axes=axes)

    # determine significant bits variable - is important that Xenium explorer correctly distinguishes between 8 bit and 16 bit
    if first_image.dtype == np.dtype('uint8'):
        significant_bits = 8
    else:
        significant_bits = 16

    file = Path(file)
    if file.exists():
        if overwrite:
            file.unlink() # delete file
        else:
            raise FileExistsError("Output file exists already ({}).\nFor overwriting it, select `overwrite=True`".format(file))

    # create metadata
    if pixelsize != 1:
        metadata = {
            **metadata,
            **{
                'PhysicalSizeX': pixelsize,
                'PhysicalSizeY': pixelsize
            }
        }
    if pixelunit is not None:
        metadata = {
            **metadata,
            **{
                'PhysicalSizeXUnit': pixelunit,
                'PhysicalSizeYUnit': pixelunit
            }
        }
    if (significant_bits is not None) & ("SignificantBits" not in metadata.keys()):
        metadata = {
            **metadata,
            **{
                'SignificantBits': significant_bits
            }
        }


    with TiffWriter(file, bigtiff=True) as tif:
        options = dict(
            photometric=photometric,
            tile=tile,
            compression=compression,
            resolutionunit='CENTIMETER',
        )
        tif.write(
            image_pyramid[0],
            subifds=subresolutions,
            resolution=(1e4 / pixelsize, 1e4 / pixelsize),
            metadata=metadata,
            **options
        )

        scale = 1
        for i in range(1, subresolutions+1):
            img = image_pyramid[i]
            #scale /= subres_steps
            #img = resize_image(img, scale_factor=1/subres_steps, axes=axes)
            tif.write(
                img,
                subfiletype=1,
                resolution=(1e4 / scale / pixelsize,1e4 / scale / pixelsize),
                **options
            )

def read_zarr_pyramid(dirstore, persist):
    # get components of zip store
    components = dirstore.listdir()

    if ".zarray" in components:
        # the store is an array which can be opened
        if persist:
            img = da.from_zarr(dirstore).persist()
        else:
            img = da.from_zarr(dirstore)
    else:
        subres = sorted([elem for elem in components if not elem.startswith(".")])
        img = []
        for s in subres:
            if persist:
                img.append(
                    da.from_zarr(dirstore, component=s).persist()
                            )
            else:
                img.append(
                    da.from_zarr(dirstore, component=s)
                            )

    return img

def read_ome_tiff(
    path,
    levels: Optional[Union[List[int], int]] = None,
    new_method: bool = True
    ):
    '''
    Function to load pyramid from `ome.tiff` file.
    From: https://www.youtube.com/watch?v=8TlAAZcJnvA
    Another good resource from 10x: https://www.10xgenomics.com/support/software/xenium-onboard-analysis/latest/analysis/xoa-output-understanding-outputs

    Args:
        path (str): The file path to the `ome.tiff` file.
        levels (Optional[Union[List[int], int]]): A list of integers representing the levels of the pyramid to load. If None, all levels are loaded. Default is None.
        new_method (bool): Is now the default method and uses a strategy found here: https://www.10xgenomics.com/support/software/xenium-onboard-analysis/latest/analysis/xoa-output-understanding-outputs.

    Returns:
        List[dask.array.Array] or dask.array.Array: The pyramid or a single level of the pyramid, represented as Dask arrays.

    '''
    if new_method:
        pyramid = []
        l = 0
        while True:
            try:
                store = imread(path, aszarr=True, level=l, is_ome=False)
                pyramid.append(da.from_zarr(store))
                l+=1 # count up
            except IndexError:
                break

    else:
        # read store
        store = imread(path, aszarr=True)

        # Open store (root group)
        grp = zarr.open(store, mode='r')

        # Read multiscale metadata
        datasets = grp.attrs["multiscales"][0]["datasets"]

        if levels is None:
            levels = range(0, len(datasets))
        # make sure level is a list
        levels = convert_to_list(levels)

        # extract images as pyramid list
        pyramid = [
            da.from_zarr(store, component=datasets[l]["path"])
            for l in levels
        ]

    # if pyramid has only one element, return only this image
    if len(pyramid) == 1:
        pyramid = pyramid[0]

    return pyramid


import math

import numpy as np
import pandas as pd
from matplotlib.gridspec import GridSpec
from matplotlib.lines import Line2D
from pandas.api.types import is_numeric_dtype

import insitupy._core.config as config
from insitupy import WITH_NAPARI
from insitupy._constants import POINTS_SYMBOL

if WITH_NAPARI:
    import napari


# geometry widget
def _update_keys_based_on_geom_type(widget, xdata):
    # retrieve current value
    current_geom_type = widget.geom_type.value
    current_key = widget.key.value

    # get either regions or annotations object
    geom_data = getattr(xdata, current_geom_type.lower())
    widget.key.choices = sorted(geom_data.metadata.keys(), key=str.casefold)

def _update_classes_on_key_change(widget, xdata):
    # get current values for geom_type and key
    current_geom_type = widget.geom_type.value
    current_key = widget.key.value

    # get either regions or annotations object
    geom_data = getattr(xdata, current_geom_type.lower())

    # update annot_class choices
    widget.annot_class.choices = ["all"] + sorted(geom_data.metadata[current_key]['classes'])

def _set_show_names_based_on_geom_type(widget):
    # retrieve current value
    current_geom_type = widget.geom_type.value

    # set the show_names tick box
    if current_geom_type == "Annotations":
        widget.show_names.value = False

    if current_geom_type == "Regions":
        widget.show_names.value = True


# Function to update the legend
def _update_categorical_legend(static_canvas, mapping, label, max_rows: int = 6):

    # Calculate the number of columns needed
    num_items = len(mapping)
    ncols = math.ceil(num_items / max_rows)

    static_canvas.figure.clear()  # Clear the current figure
    axes = static_canvas.figure.subplots()  # Create new axes
    legend_handles = [Line2D([0], [0],
                             marker='o', color='w', label=n,
                             markerfacecolor=c, markeredgecolor='k',
                             markersize=7) for n,c in mapping.items()]
    axes.legend(handles=legend_handles, loc="center", title=label, ncols=ncols,
                fontsize=8, title_fontsize=10,
                labelspacing=0.7, borderpad=0.5)
    axes.set_axis_off()
    static_canvas.draw()  # Redraw the canvas

def _update_continuous_legend(static_canvas, mapping, label):
    static_canvas.figure.clear()  # Clear the current figure
    gs = GridSpec(1, 1, top=1.2, bottom=0.6, left=-0.5, right=1.5)  # Define the grid spec
    axes = static_canvas.figure.add_subplot(gs[0])  # Add subplot with the grid spec

    colorbar = static_canvas.figure.colorbar(mapping, ax=axes, orientation='horizontal')
    colorbar.set_label(label, fontsize=10)
    colorbar.ax.tick_params(labelsize=8)  # Adjust tick label size
    #colorbar.set_ticks(np.linspace(norm.vmin, norm.vmax, num=5))  # Set the number of ticks
    axes.set_axis_off()
    static_canvas.draw()  # Redraw the canvas

def _update_colorlegend():

    # # automatically get layer
    # candidate_layers = [l for l in config.viewer.layers if l.name.startswith(f"{config.current_data_name}")]
    # try:
    #     # always choose the candidate layer that is on top
    #     layer_name = candidate_layers[-1].name
    # except IndexError:
    #     raise ValueError("No layer with cellular transcriptomic data found. First add a layer using the 'Show Data' widget.")

    # # extract layer
    # layer = config.viewer.layers[layer_name]

    layer = config.viewer.layers.selection.active

    if isinstance(layer, napari.layers.points.points.Points):
        # get values
        values = layer.properties["value"]
        color_values = layer.face_color


        from insitupy.plotting._colors import continuous_data_to_rgba
        if is_numeric_dtype(values):
            rgba_list, mapping = continuous_data_to_rgba(data=values,
                                    cmap=layer.face_colormap.name,
                                    #upper_climit_pct=upper_climit_pct,
                                    return_mapping=True
                                    )

            _update_continuous_legend(static_canvas=config.static_canvas,
                                    mapping=mapping,
                                    label=layer.name)

        else:
            # substitute pd.NA with np.nan
            values = pd.Series(values).fillna(np.nan).values
            # assume the data is categorical
            #mapping = {category: tuple(rgba) for category, rgba in zip(values, color_values)}
            unique_values = list(set(values))
            mapping = {str(v): tuple(color_values[list(values).index(v)]) for v in unique_values}
            # sort mapping dict
            mapping = {elem: mapping[elem] for elem in sorted(mapping.keys())}

            _update_categorical_legend(static_canvas=config.static_canvas,
                                    mapping=mapping, label=layer.name)

        # # create color mapping
        # rgba_list, mapping = _data_to_rgba(values, return_mapping=True)

        # if isinstance(mapping, dict):
        #     _update_categorical_legend(static_canvas=config.static_canvas,
        #                                mapping=mapping, label=layer.name)
        # else:
        #     _update_continuous_legend(static_canvas=config.static_canvas,
        #                               mapping=mapping,
        #                               label=layer.name)


def _refresh_widgets_after_data_change(xdata, points_widget, boundaries_widget, filter_widget):
    config.init_viewer_config(xdata)

    # set choices
    boundaries_widget.key.choices = config.masks

    # reset the currently selected key to None
    points_widget.value.value = None

    # add last addition to recent
    points_widget.recent.choices = sorted(config.recent_selections)
    points_widget.recent.value = None

    # update obs in filter widget
    filter_widget.obs_key.choices = config.value_dict["obs"]

    # set only the last cell layer visible
    cell_layers = []
    for elem in xdata.viewer.layers:
        if isinstance(elem, napari.layers.points.points.Points):
            if not elem.name.startswith(POINTS_SYMBOL):
                # only if the layer is not a point annotation layer, it is added
                cell_layers.append(elem)
    #point_layers = [elem for elem in xdata.viewer.layers if isinstance(elem, napari.layers.points.points.Points)]
    n_cell_layers = len(cell_layers)

    # # make only last cell layer visible
    # for i, l in enumerate(cell_layers):
    #     if i < n_cell_layers-1:
    #         l.visible = False


import os
from numbers import Number
from pathlib import Path
from typing import List, Union

import matplotlib.pyplot as plt
import pandas as pd
import seaborn as sns

from ..io.files import read_json
from ..io.plots import save_and_show_figure
from .utils import get_nrows_maxcols


def find_xenium_outputs(
    path: Union[str, os.PathLike, Path],
    startswith: str = 'output-XET'
    ) -> List:
    print(f"Searching for directories starting with '{startswith}' in {str(path)}")
    search_results = []
    for root, dirs, files in os.walk(path):
        for d in dirs:
            if d.startswith(startswith):
                search_results.append(os.path.join(root, d))

    print(f"Found {len(search_results)} Xenium output directories.")
    return search_results

def collect_qc_data(
    data_folders: List[Union[str, os.PathLike, Path]]
    ) -> pd.DataFrame:

    cats = ["run_name", "slide_id", "region_name", "preservation_method",
            "num_cells", "transcripts_per_cell",
            "transcripts_per_100um", "panel_organism", "panel_tissue_type"]

    results = []
    for f in data_folders:
        metadata = read_json(Path(f) / "experiment.xenium")
        extracted = [metadata[c] for c in cats]
        results.append(extracted)

    data = pd.DataFrame(results, columns=cats)
    return data

def plot_qc(
    data: pd.DataFrame,
    x: str = "preservation_method",
    cats: List[str] = ["num_cells", "transcripts_per_cell", "transcripts_per_100um"],
    max_cols: int = 4,
    fontsize: int = 22,
    size: Number = 10,
    savepath: Union[str, os.PathLike, Path] = None,
    save_only: bool = False,
    dpi_save: int = 300
    ):
    # plot
    plt.rcParams.update({'font.size': fontsize})
    n_plots, nrows, ncols = get_nrows_maxcols(len(cats), max_cols=max_cols)
    fig, axs = plt.subplots(nrows, ncols, figsize=(9*ncols, 8*nrows))

    if n_plots > 1:
        axs = axs.ravel()
    else:
        axs = [axs]

    for i, cat in enumerate(cats):
        sns.boxplot(data=data, x=x, y=cat, color="w",
                    boxprops={"facecolor": 'w'}, fliersize=0,
                    ax=axs[i], )
        sns.stripplot(data=data,
                      x=x, y=cat,
                      hue="panel_tissue_type",
                      size=size,
                      ax=axs[i]
                      )
        axs[i].set_title(cat)
        axs[i].set_ylabel(None)

        if i+1 == ncols:
            # move legend out of the plot
            axs[i].legend(loc='center left', bbox_to_anchor=(1, 0.5))
        else:
            # remove legend
            axs[i].get_legend().remove()

    save_and_show_figure(savepath=savepath, fig=fig, save_only=save_only, dpi_save=dpi_save, tight=True)



import json
import os
import shutil
from pathlib import Path
from typing import Union

from insitupy.utils.utils import nested_dict_numpy_to_list


def read_json(
    file: Union[str, os.PathLike, Path],
    ) -> dict:
    '''
    Function to load json files as dictionary.
    '''
    # load metadata file
    with open(file, "r") as metafile:
        metadata = json.load(metafile)

    return metadata


def write_dict_to_json(
    dictionary: dict,
    file: Union[str, os.PathLike, Path],
    ):
    try:
        dict_json = json.dumps(dictionary, indent=4)
        with open(file, "w") as metafile:
                metafile.write(dict_json)
    except TypeError:
        # one reason for this type error could be that there are ndarrays in the dict
        # convert them to lists
        nested_dict_numpy_to_list(dictionary)

        dict_json = json.dumps(dictionary, indent=4)
        with open(file, "w") as metafile:
                metafile.write(dict_json)


def check_overwrite_and_remove_if_true(
    path: Union[str, os.PathLike, Path],
    overwrite: bool = False
    ):
    path = Path(path)
    if path.exists():
        if overwrite:
            if path.is_dir():
                shutil.rmtree(path) # delete directory
            elif path.is_file():
                path.unlink() # delete file
            else:
                raise ValueError(f"Path is neither a directory nor a file. What is it? {str(path)}")
        else:
            raise FileExistsError(f"The output file already exists at {path}. To overwrite it, please set the `overwrite` parameter to True."
)


from .datasets import (human_brain_cancer, human_breast_cancer,
                       human_kidney_nondiseased, human_lung_cancer,
                       human_lymph_node, human_lymph_node_5k,
                       human_pancreatic_cancer, human_skin_melanoma,
                       list_downloaded_datasets, xenium_test_dataset)
from .download import download_url



from .expression_along_axis import (cell_abundance_along_axis,
                                    cell_expression_along_axis,
                                    expr_along_obs_val)
from .plots import plot_cellular_composition, plot_colorlegend
from .volcano import volcano_plot



from typing import Literal

import anndata
import numpy as np
from scipy.sparse import issparse


# checker functions for data sanity
def check_adata(adata):
    if type(adata) is not anndata.AnnData:
        raise TypeError('Input is not a valid AnnData object')


def check_batch(batch, obs, verbose=False):
    if batch not in obs:
        raise ValueError(f'column {batch} is not in obs')
    elif verbose:
        print(f'Object contains {obs[batch].nunique()} batches.')


def check_hvg(hvg, hvg_key, adata_var):
    if type(hvg) is not list:
        raise TypeError('HVG list is not a list')
    else:
        if not all(i in adata_var.index for i in hvg):
            raise ValueError('Not all HVGs are in the adata object')
    if not hvg_key in adata_var:
        raise KeyError('`hvg_key` not found in `adata.var`')

def check_sanity(adata, batch, hvg, hvg_key):
    check_adata(adata)
    check_batch(batch, adata.obs)
    if hvg:
        check_hvg(hvg, hvg_key, adata.var)


def check_integer_counts(X):
    '''
    Check if a matrix consists of raw integer counts or if it is processed already.
    '''

    # convert sparse matrix to numpy array
    if issparse(X):
        X = X.toarray()

    # check if the matrix contains raw counts
    if not np.all(np.modf(X)[0] == 0):
        raise ValueError("Anndata object does not contain raw counts. Preprocessing aborted.")


def is_integer_counts(X):
    '''
    Check if a matrix consists of raw integer counts or if it is processed already.
    '''

    # convert sparse matrix to numpy array
    if issparse(X):
        X = X.toarray()

    # check if the matrix contains raw counts
    return np.all(np.modf(X)[0] == 0)

def check_raw(adata, use_raw, layer=None):
    # check if plotting raw data
    if use_raw:
        adata_X = adata.raw.X
        adata_var = adata.raw.var
        adata_var_names = adata.raw.var_names
    else:
        if layer is None:
            adata_X = adata.X
        else:
            #adata_X = adata.layers[layer].toarray()
            adata_X = adata.layers[layer]

        if issparse(adata_X):
            adata_X = adata_X.toarray()

        adata_var = adata.var
        adata_var_names = adata.var_names

    return adata_X, adata_var, adata_var_names

def check_zip(path):
    # check if the output directory is going to be zipped or not
    if path.suffix == ".zip":
        zip_output = True
        path = path.with_suffix("")
    elif path.suffix == "":
        zip_output = False
    else:
        raise ValueError(f"The specified output path ({path}) must be a valid directory or a zip file. It does not need to exist yet.")

    return zip_output

# Function to check if there are any valid labels in matplotlib figure
def has_valid_labels(ax):
    for artist in ax.get_legend_handles_labels()[0]:  # Get the handles (artists)
        if artist.get_label() and not artist.get_label().startswith('_'):
            return True
    return False

def is_valid_rgb_tuple(value):
    """
    Check if a value is a valid RGB tuple.

    A valid RGB tuple is defined as a list or tuple containing three integers,
    each in the range of 0 to 255.

    Parameters:
    value (list or tuple): The value to check.

    Returns:
    bool: True if the value is a valid RGB tuple, False otherwise.
    """
    if isinstance(value, (list, tuple)) and len(value) == 3:
        return all(isinstance(v, int) and 0 <= v <= 255 for v in value)
    return False

def check_rgb_column(df, column_name):
    """
    Check if a specified column in a DataFrame contains only valid RGB tuples.

    This function checks if the specified column exists in the DataFrame and
    verifies that all entries in the column are valid RGB tuples.

    Parameters:
    df (pd.DataFrame): The DataFrame to check.
    column_name (str): The name of the column to validate.

    Returns:
    bool: True if all values in the column are valid RGB tuples, False otherwise.

    Raises:
    ValueError: If the specified column does not exist in the DataFrame.
    """
    if column_name not in df.columns:
        raise ValueError(f"Column '{column_name}' does not exist in the DataFrame.")

    # Check if all values in the specified column are valid RGB tuples
    return df[column_name].apply(is_valid_rgb_tuple).all()

def _check_assignment(
    data,
    key: str,
    modality: Literal["annotations", "regions"],
    force_assignment: bool = False,
    verbose: bool = False
):
    try:
        column = data.cells.matrix.obsm[modality].columns
    except KeyError:
        do_assignment = True
    else:
        if key in column:
            do_assignment = False
        else:
            do_assignment = True

    if do_assignment or force_assignment:
        if modality == "annotations":
            # assign annotations
            data.assign_annotations(keys=key)
        elif modality == "regions":
            # assign regions
            data.assign_regions(keys=key)
    else:
        if verbose:
            print(f"{modality.capitalize()} with key '{key}' have already been assigned to the dataset.")



import os
import textwrap
from pathlib import Path
from typing import Literal, Optional, Union

import matplotlib.pyplot as plt
import pandas as pd
from matplotlib.lines import Line2D
from napari.viewer import Viewer

import insitupy._core.config as config
from insitupy._constants import DEFAULT_CATEGORICAL_CMAP
from insitupy._core._checks import _check_assignment
from insitupy.io.plots import save_and_show_figure
from insitupy.plotting._colors import _data_to_rgba
from insitupy.utils.utils import get_nrows_maxcols


def plot_colorlegend(
    viewer: Viewer,
    layer_name: Optional[str] = None,
    savepath: Union[str, os.PathLike, Path] = None,
    save_only: bool = False,
    dpi_save: int = 300,
    ):
    # automatically get layer
    if layer_name is None:
        candidate_layers = [l for l in viewer.layers if l.name.startswith(f"{config.current_data_name}")]
        try:
            layer_name = candidate_layers[0].name
        except IndexError:
            raise ValueError("No layer with cellular transcriptomic data found. First add a layer using the 'Show Data' widget.")

    # extract layer
    layer = viewer.layers[layer_name]

    # get values
    values = layer.properties["value"]

    # create color mapping
    rgba_list, mapping, cmap = _data_to_rgba(values)

    if isinstance(mapping, dict):
        # categorical colorbar
        # create a figure for the colorbar
        fig, ax = plt.subplots(
            #figsize=(5, 3)
            )
        fig.subplots_adjust(bottom=0.5)

        circles = [Line2D([0], [0],
                            marker='o', color='w', label=label,
                            markerfacecolor=color, markeredgecolor='k', markersize=15) for label, color in mapping.items()]

        ax.legend(handles=circles, loc="center", labelspacing=1, borderpad=0.5)
        ax.set_axis_off()

    else:
        # continuous colorlegend
        # create a figure for the colorbar
        fig, ax = plt.subplots(
            figsize=(6, 1)
            )
        fig.subplots_adjust(bottom=0.5)

        # Add the colorbar to the figure
        cbar = fig.colorbar(mapping, orientation='horizontal', cax=ax)
        cbar.ax.set_title(layer_name)

    save_and_show_figure(savepath=savepath, fig=fig, save_only=save_only, dpi_save=dpi_save, tight=False)
    plt.show()


def plot_cellular_composition(
    data,
    cell_type_col: str,
    key: str,
    modality: Literal["regions", "annotations"] = "regions",
    plot_type: Literal["pie", "bar", "barh"] = "barh",
    force_assignment: bool = False,
    max_cols: int = 4,
    savepath: Union[str, os.PathLike, Path] = None,
    show_labels: bool = False,
    adjust_labels: bool = False,
    label_threshold: float = 2.,
    return_data: bool = False,
    save_only: bool = False,
    dpi_save: int = 300,
    ):

    """
    Plots the composition of cell types for specified regions or annotations.

    This function generates pie charts or a single stacked bar plot to visualize the proportions of different cell types
    within specified regions or annotations. It can optionally save the plot to a file and
    return the composition data.

    Args:
        data: The dataset containing cell information.
        cell_type_col (str): The column name in `adata.obs` that contains cell type information.
        key (str): The key to access the specific annotation or region in `adata.obsm`.
        modality (Literal["regions", "annotations"], optional): The modality to use, either "regions" or "annotations". Default is "regions".
        plot_type (Literal["pie", "bar"], optional): The type of plot to generate, either "pie" or "bar". Default is "pie".
        force_assignment (bool, optional): If True, forces reassignment of cells to the requested annotation. Default is False.
        max_cols (int, optional): Maximum number of columns for subplots. Defaults to 4.
        savepath (Union[str, os.PathLike, Path], optional): The path to save the plot. If None, the plot is not saved. Default is None.
        show_labels (bool, optional): If True, displays percentage labels on the pie charts. Default is False.
        adjust_labels (bool, optional): If True, adjusts the labels to avoid overlap. Default is False.
        label_threshold (float, optional): The threshold percentage above which labels are displayed. Default is 2.0.
        return_data (bool, optional): If True, returns the composition data as a DataFrame. Default is False.
        save_only (bool, optional): If True, only saves the plot without displaying it. Default is False.
        dpi_save (int, optional): The resolution in dots per inch for the saved plot. Default is 300.

    Returns:
        pd.DataFrame: A DataFrame containing the composition of cell types if `return_data` is True.

    Raises:
        ValueError: If the specified key or modality is not found in the data.

    Example:
        >>> compositions = plot_cell_composition(data, cell_type_col="cell_type", key="region_1", plot_type="bar", return_data=True)
        >>> print(compositions)
    """
    if adjust_labels:
        try:
            from adjustText import adjust_text
        except ImportError:
            raise ImportError("The 'adjustText' module is required for label adjustment. Please install it with `pip install adjusttext` or select adjust_labels=False.")

    # check whether the cells were already assigned to the requested annotation
    _check_assignment(data=data, key=key, force_assignment=force_assignment, modality=modality)

    # retrieve data
    adata = data.cells.matrix
    assignment_series = adata.obsm[modality][key]
    cats = sorted([elem for elem in assignment_series.unique() if (elem != "unassigned") & ("&" not in elem)])

    # calculate compositions
    compositions = {}
    for cat in cats:
        idx = assignment_series[assignment_series == cat].index
        compositions[cat] = adata.obs[cell_type_col].loc[idx].value_counts(normalize=True) * 100 # calculate percentage
    compositions = pd.DataFrame(compositions)

    # Define a function to display percentages above the threshold
    def autopct_func(pct):
        return ('%1.1f%%' % pct) if pct > label_threshold else ''

    if plot_type == "pie":
        # Plot pie charts for each area
        n_plots, nrows, ncols = get_nrows_maxcols(len(cats), max_cols)
        fig, axs = plt.subplots(nrows, ncols, figsize=(6*ncols,6*nrows))

        if n_plots > 1:
            axs = axs.ravel()
        else:
            axs = [axs]

        for i, area in enumerate(compositions.columns):
            if show_labels:
                wedges, texts, autotexts = axs[i].pie(compositions[area],
                                                    autopct=autopct_func, pctdistance=1.15,
                                                    colors=DEFAULT_CATEGORICAL_CMAP.colors
                                                    )
            else:
                wedges, texts = axs[i].pie(compositions[area],
                                                    colors=DEFAULT_CATEGORICAL_CMAP.colors
                                                    )

            title_str = textwrap.fill(f'Proportions of Cell Types in {area}', width=20)
            axs[i].set_title(title_str)

            if adjust_labels:
                # Adjust text to avoid overlap
                adjust_text(texts + autotexts, ax=axs[i], arrowprops=dict(arrowstyle="->", color='k', lw=0.5))

        # Add a legend
        fig.legend(wedges, compositions.index, loc='center left', bbox_to_anchor=(0.92, 0.5))

    elif plot_type in ["bar", "barh"]:
        # Plot a single stacked bar plot
        if plot_type == "bar":
            fig_width = 1*len(cats)
            fig_height = 6
            ylabel = "%"
            xlabel = modality
        else:
            fig_width = 6
            fig_height = 1*len(cats)
            ylabel = modality
            xlabel = "%"
        compositions.T.plot(kind=plot_type, stacked=True, figsize=(fig_width, fig_height),
                            width=0.7,
                            color=DEFAULT_CATEGORICAL_CMAP.colors)
        plt.gca().invert_yaxis()
        plt.title('Cell type composition')
        plt.ylabel(ylabel)
        plt.xlabel(xlabel)
        plt.legend(title='Cell Types', bbox_to_anchor=(1.05, 1), loc='upper left')

    save_and_show_figure(savepath=savepath, fig=plt.gcf(), save_only=save_only, dpi_save=dpi_save, tight=False)

    plt.tight_layout()
    plt.show()

    if return_data:
        return compositions



import gc
import os
from pathlib import Path
from typing import List, Literal, Optional, Union

import numpy as np
from dask_image.imread import imread
from parse import *

from insitupy import __version__
from insitupy._core.insitudata import InSituData
from insitupy._exceptions import UnknownOptionError
from insitupy.images import (ImageRegistration, clip_image_histogram,
                             deconvolve_he, resize_image)
from insitupy.images.utils import otsu_thresholding
from insitupy.io.files import write_dict_to_json
from insitupy.utils.utils import convert_to_list
from insitupy.utils.utils import textformat as tf


def register_images(
    data: InSituData,
    image_to_be_registered: Union[str, os.PathLike, Path],
    image_type: Literal["histo", "IF"],
    channel_names: Union[str, List[str]],
    channel_name_for_registration: Optional[str] = None,  # name used for the nuclei image. Only required for IF images.
    template_image_name: str = "nuclei",
    save_registered_images: bool = True,
    output_dir: Union[str, os.PathLike, Path] = None,
    min_good_matches_per_area: int = 5, # unit: 1/mm²
    test_flipping: bool = True,
    decon_scale_factor: float = 0.2,
    physicalsize: str = 'µm',
    prefix: str = "",
    ):
    """
    Register images stored in an InSituData object.

    Args:
        data (InSituData): The InSituData object containing the images.
        image_to_be_registered (Union[str, os.PathLike, Path]): Path to the image to be registered.
        image_type (Literal["histo", "IF"]): Type of the image, either "histo" or "IF".
        channel_names (Union[str, List[str]]): Names of the channels in the image.
        channel_name_for_registration (Optional[str], optional): Name of the channel used for registration. Required for IF images. Defaults to None.
        template_image_name (str, optional): Name of the template image. Defaults to "nuclei".
        save_registered_images (bool, optional): Whether to save the registered images. Defaults to True.
        min_good_matches (int, optional): Minimum number of good matches required for registration. Defaults to 20.
        test_flipping (bool): Whether to test flipping of images during registration. Defaults to True.
        decon_scale_factor (float, optional): Scale factor for deconvolution. Defaults to 0.2.
        physicalsize (str, optional): Unit of physical size. Defaults to 'µm'.
        prefix (str, optional): Prefix for the output files. Defaults to "".

    Raises:
        ValueError: If `image_type` is "IF" and `channel_name_for_registration` is None.
        FileNotFoundError: If the image to be registered is not found.
        ValueError: If more than one image name is retrieved for histo images.
        ValueError: If no image name is found in the file.
        UnknownOptionError: If an unknown image type is provided.
        TypeError: If `channel_name_for_registration` is None for IF images.
        ValueError: If no channel indicator `C` is found in the image axes.

    Returns:
        None
    """

    # if image type is IF, the channel name for registration needs to be given
    if image_type == "IF" and channel_name_for_registration is None:
        raise ValueError(f'If `image_type" is "IF", `channel_name_for_registration is not allowed to be `None`.')

    if output_dir is None:
        # define output directory
        output_dir = data.path.parent / "registered_images"
    else:
        output_dir = Path(output_dir) / "registered_images"
        output_dir.mkdir(parents=True, exist_ok=True)

    # if output_dir.is_dir() and not force:
    #     raise FileExistsError(f"Output directory {output_dir} exists already. If you still want to run the registration, set `force=True`.")

    # check if image path exists
    image_to_be_registered = Path(image_to_be_registered)
    if not image_to_be_registered.is_file():
        raise FileNotFoundError(f"No such file found: {str(image_to_be_registered)}")

    # make sure the given image names are in a list
    channel_names = convert_to_list(channel_names)

    # determine the structure of the image axes and check other things
    axes_template = "YX"
    if image_type == "histo":
        axes_image = "YXS"

        # make sure that there is only one image name given
        if len(channel_names) > 1:
            raise ValueError(f"More than one image name retrieved ({channel_names})")

        if len(channel_names) == 0:
            raise ValueError(f"No image name found in file {image_to_be_registered}")

    elif image_type == "IF":
        axes_image = "CYX"
    else:
        raise UnknownOptionError(image_type, available=["histo", "IF"])

    print(f'\tProcessing following {image_type} images: {tf.Bold}{", ".join(channel_names)}{tf.ResetAll}', flush=True)

    # read images
    print("\t\tLoading images to be registered...", flush=True)
    image = imread(image_to_be_registered) # e.g. HE image

    # sometimes images are read with an empty time dimension in the first axis.
    # If this is the case, it is removed here.
    if len(image.shape) == 4:
        image = image[0]

    # # read images in InSituData object
    template = data.images[template_image_name][0] # usually the nuclei/DAPI image is the template. Use highest resolution of pyramid.

    # extract OME metadata
    #ome_metadata_template = data.images.metadata[template_image_name]["OME"]

    # get pixel size from image metadata
    pixel_size = data.images.metadata[template_image_name]["pixel_size"]

    # extract pixel size for x and y from OME metadata
    #pixelsizes = {key: ome_metadata_template['Image']['Pixels'][key] for key in ['PhysicalSizeX', 'PhysicalSizeY']}

    # generate OME metadata for saving
    ome_metadata = {
        'SignificantBits': 8,
        'PhysicalSizeXUnit': physicalsize,
        'PhysicalSizeYUnit': physicalsize,
        'PhysicalSizeX': pixel_size,
        'PhysicalSizeY': pixel_size
        }

    # determine minimum number of good matches that are necessary for the registration to be performed
    h, w = template.shape[:2]
    image_area = h * w * pixel_size**2 / 1000**2 # in mm²
    min_good_matches = int(min_good_matches_per_area * image_area)

    # the selected image will be a grayscale image in both cases (nuclei image or deconvolved hematoxylin staining)
    #axes_selected = "YX"
    if image_type == "histo":
        print("\t\tRun color deconvolution", flush=True)
        # deconvolve HE - performed on resized image to save memory
        # TODO: Scale to max width instead of using a fixed scale factor before deconvolution (`scale_to_max_width`)
        nuclei_img, eo, dab = deconvolve_he(img=resize_image(image, scale_factor=decon_scale_factor, axes="YXS"),
                                    return_type="grayscale", convert=True)

        # bring back to original size
        nuclei_img = resize_image(nuclei_img, scale_factor=1/decon_scale_factor, axes="YX")

        # set nuclei_channel and nuclei_axis to None
        channel_name_for_registration = channel_axis = None
    else:
        # image_type is "IF" then
        # get index of nuclei channel
        channel_id_for_registration = channel_names.index(channel_name_for_registration)
        channel_axis = axes_image.find("C")

        if channel_axis == -1:
            raise ValueError(f"No channel indicator `C` found in image axes ({axes_image})")

        print(f"\t\tSelect image with nuclei from IF image (channel index: {channel_id_for_registration})", flush=True)
        # # select nuclei channel from IF image
        # if channel_name_for_registration is None:
        #     raise TypeError("Argument `nuclei_channel` should be an integer and not NoneType.")

        # select dapi channel for registration and convert to numpy array
        nuclei_img = np.take(image, channel_id_for_registration, channel_axis).compute()

    # Setup image registration objects - is important to load and scale the images.
    # The reason for this are limits in C++, not allowing to perform certain OpenCV functions on big images.

    # First: Setup the ImageRegistration object for the whole image (before deconvolution in histo images and multi-channel in IF)
    imreg_complete = ImageRegistration(
        image=image,
        template=template,
        axes_image=axes_image,
        axes_template=axes_template,
        verbose=True
        )
    # load and scale the whole image
    print('Load and scale image data containing all channels.')
    imreg_complete.load_and_scale_images()

    # setup ImageRegistration object with the nucleus image (either from deconvolution or just selected from IF image)
    axes_selected = "YX"
    imreg_selected = ImageRegistration(
        image=nuclei_img,
        template=imreg_complete.template,
        axes_image=axes_selected,
        axes_template=axes_template,
        max_width=4000,
        convert_to_grayscale=False,
        perspective_transform=False,
        min_good_matches=min_good_matches
    )

    # run all steps to extract features and get transformation matrix
    print('Load and scale image data containing only the channels required for registration.')
    imreg_selected.load_and_scale_images()

    print("\t\tExtract common features from image and template", flush=True)
    # perform registration to extract the common features ptsA and ptsB
    imreg_selected.extract_features(test_flipping=test_flipping)
    imreg_selected.calculate_transformation_matrix()

    if image_type == "histo":
        # in case of histo RGB images, the channels are in the third axis and OpenCV can transform them
        if imreg_complete.image_resized is None:
            imreg_selected.image = imreg_complete.image  # use original image
        else:
            imreg_selected.image_resized = imreg_complete.image_resized  # use resized original image

        # perform registration
        imreg_selected.perform_registration()

        if save_registered_images:
            # save files
            identifier = f"{prefix}__{data.slide_id}__{data.sample_id}__{channel_names[0]}"
            imreg_selected.save(
                output_dir=output_dir,
                identifier = identifier,
                axes=axes_image,
                photometric='rgb',
                ome_metadata=ome_metadata
                )

            # # save metadata
            # data.metadata["method_params"]['images'][f'registered_{channel_names[0]}_filepath'] = os.path.relpath(imreg_selected.outfile, data.path).replace("\\", "/")
            # write_dict_to_json(data.metadata["method_params"], data.path / "experiment_modified.xenium")
            # #self._save_metadata_after_registration()

        data.images.add_image(
            image=imreg_selected.registered,
            name=channel_names[0],
            axes=axes_image,
            pixel_size=pixel_size,
            ome_meta=ome_metadata,
            overwrite=True
            )

        del imreg_complete, imreg_selected, image, template, nuclei_img, eo, dab
    else:
        # image_type is IF
        # In case of IF images the channels are normally in the first axis and each channel is registered separately
        # Further, each channel is then saved separately as grayscale image.

        # iterate over channels
        for i, n in enumerate(channel_names):
            # skip the DAPI image
            if n == channel_name_for_registration:
                break

            if imreg_complete.image_resized is None:
                # select one channel from non-resized original image
                imreg_selected.image = np.take(imreg_complete.image, i, channel_axis)
            else:
                # select one channel from resized original image
                imreg_selected.image_resized = np.take(imreg_complete.image_resized, i, channel_axis)

            # perform registration
            imreg_selected.perform_registration()

            if save_registered_images:
                # save files
                identifier = f"{data.slide_id}__{data.sample_id}__{n}"

                imreg_selected.save(
                    output_dir=output_dir,
                    identifier=identifier,
                    axes='YX',
                    photometric='minisblack',
                    ome_metadata=ome_metadata
                    )

                # # save metadata
                # data.metadata["method_params"]['images'][f'registered_{n}_filepath'] = os.path.relpath(imreg_selected.outfile, data.path).replace("\\", "/")
                # write_dict_to_json(data.metadata["method_params"], data.path / "experiment_modified.xenium")
                # #self._save_metadata_after_registration()
            # if add_registered_image:
            data.images.add_image(
                image=imreg_selected.registered,
                name=n,
                axes=axes_image,
                pixel_size=pixel_size,
                ome_meta=ome_metadata,
                overwrite=True
                )

        # free RAM
        del imreg_complete, imreg_selected, image, template, nuclei_img
    gc.collect()


from .axes import ImageAxes
from .io import *
from .registration import ImageRegistration
from .utils import (clip_image_histogram, convert_to_8bit_func, deconvolve_he,
                    fit_image_to_size_limit, resize_image, scale_to_max_width)



from numbers import Number
from typing import List, Optional
from warnings import warn

import matplotlib
import numpy as np
import pandas as pd
from matplotlib.colors import rgb2hex
from shapely.geometry.multipolygon import MultiPolygon

import insitupy._core._callbacks
import insitupy._core.config as config
from insitupy import WITH_NAPARI
from insitupy._core._layers import _create_points_layer, _update_points_layer

from .._constants import (ANNOTATIONS_SYMBOL, POINTS_SYMBOL, REGION_CMAP,
                          REGIONS_SYMBOL)
from ..images.utils import create_img_pyramid
from ._callbacks import (_refresh_widgets_after_data_change,
                         _set_show_names_based_on_geom_type,
                         _update_classes_on_key_change, _update_colorlegend,
                         _update_keys_based_on_geom_type)

if WITH_NAPARI:
    import napari
    from magicgui import magic_factory, magicgui
    from magicgui.widgets import FunctionGui

    from ._layers import _add_geometries_as_layer

    def _initialize_widgets(
        xdata # InSituData object
        ) -> List[FunctionGui]:

        # access viewer from InSituData
        viewer = xdata.viewer

        if xdata.cells is None:
            add_cells_widget = None
            move_to_cell_widget = None
            add_boundaries_widget = None,
            filter_cells_widget = None
        else:
            # initialize data_name of viewer
            config.init_data_name()
            # initialize viewer configuration
            config.init_viewer_config(xdata=xdata,
                                        #data_name=config.current_data_name
                                        )
            config.init_recent_selections()

            data_names = ["main"]
            if xdata.alt is not None:
                alt = xdata.alt
                for k in alt.keys():
                    data_names.append(k)

            @magicgui(
                call_button=False,
                data_name= {'choices': data_names, 'label': 'Dataset:'}
            )
            def select_data(
                data_name="main"
            ):
                pass

            # connect key change with update function
            @select_data.data_name.changed.connect
            def update_widgets_on_data_change(event=None):
                config.current_data_name = select_data.data_name.value
                insitupy._core._callbacks._refresh_widgets_after_data_change(xdata,
                                                    add_cells_widget,
                                                    add_boundaries_widget,
                                                    filter_cells_widget
                                                    )

            if len(config.masks) > 0:
                @magicgui(
                    call_button='Show',
                    key={'choices': config.masks, 'label': 'Masks:'}
                )
                def add_boundaries_widget(
                    key
                ):
                    layer_name = f"{config.current_data_name}-boundaries-{key}"

                    if layer_name not in viewer.layers:
                        # get geopandas dataframe with regions
                        mask = config.boundaries[key]

                        # get metadata for mask
                        metadata = config.boundaries.metadata
                        pixel_size = metadata[key]["pixel_size"]

                        if not isinstance(mask, list):
                            # generate pyramid of the mask
                            mask_pyramid = create_img_pyramid(img=mask, nsubres=6)
                        else:
                            mask_pyramid = mask

                        # add masks as labels to napari viewer
                        viewer.add_labels(mask_pyramid, name=layer_name, scale=(pixel_size,pixel_size))
                    else:
                        print(f"Layer '{layer_name}' already in layer list.", flush=True)
            else:
                add_boundaries_widget = None

            def _update_values_on_key_change(widget):
                current_key = widget.key.value
                widget.value.choices = config.value_dict[current_key]

            @magicgui(
                call_button='Add',
                key={'choices': ["genes", "obs", "obsm"], 'label': 'Key:'},
                value={'choices': config.genes, 'label': "Value:"},
                size={'label': 'Size [µm]'},
                recent={'choices': [""], 'label': "Recent:"},
                add_new_layer={'label': 'Add new layer'}
                )
            def add_cells_widget(
                key="genes",
                value=None,
                size=8,
                recent=None,
                add_new_layer=False,
                viewer=viewer
                ) -> napari.types.LayerDataTuple:

                # get names of cells
                cell_names = config.adata.obs_names.values

                #layers_to_add = []
                if value is not None or recent is not None:
                    if value is None:
                        key = recent.split(":", maxsplit=1)[0]
                        value = recent.split(":", maxsplit=1)[1]
                    #if gene not in viewer.layers:
                    # get expression values
                    if key == "genes":
                        gene_loc = config.adata.var_names.get_loc(value)
                        color_value = config.X[:, gene_loc]
                    elif key == "obs":
                        color_value = config.adata.obs[value]
                    elif key == "obsm":
                        #TODO: Implement it for obsm
                        obsm_key = value.split("#", maxsplit=1)[0]
                        obsm_col = value.split("#", maxsplit=1)[1]
                        data = config.adata.obsm[obsm_key]

                        if isinstance(data, pd.DataFrame):
                            color_value = data[obsm_col].values
                        elif isinstance(data, np.ndarray):
                            color_value = data[:, int(obsm_col)-1].values
                        else:
                            warn("Data in `obsm` needs to be either pandas DataFrame or numpy array to be parsed.")
                        pass
                    else:
                        print("Unknown key selected.", flush=True)

                    new_layer_name = f"{config.current_data_name}-{value}"

                    # get layer names from the current data
                    layer_names_for_current_data = [elem.name for elem in viewer.layers if elem.name.startswith(config.current_data_name)]

                    # select only point layers
                    layer_names_for_current_data = [elem for elem in layer_names_for_current_data if isinstance(viewer.layers[elem], napari.layers.points.points.Points)]

                    # save last addition to add it to recent in the callback
                    config.recent_selections.append(f"{key}:{value}")

                    if len(layer_names_for_current_data) == 0:

                        # create points layer for genes
                        gene_layer = _create_points_layer(
                            points=config.points,
                            color_values=color_value,
                            #name=f"{config.current_data_name}-{gene}",
                            name=new_layer_name,
                            point_names=cell_names,
                            point_size=size,
                            upper_climit_pct=99
                        )
                        return gene_layer
                        #layers_to_add.append(gene_layer)
                    else:
                        if not add_new_layer:
                            #print(f"Key '{gene}' already in layer list.", flush=True)
                            # update the existing points layer
                            layer = viewer.layers[layer_names_for_current_data[0]]
                            _update_points_layer(
                                layer=layer,
                                new_color_values=color_value,
                                new_name=new_layer_name,
                            )
                        else:
                            # create new points layer for genes
                            gene_layer = _create_points_layer(
                                points=config.points,
                                color_values=color_value,
                                #name=f"{config.current_data_name}-{gene}",
                                name=new_layer_name,
                                point_names=cell_names,
                                point_size=size,
                                upper_climit_pct=99
                            )
                            return gene_layer

            @magicgui(
                call_button='Filter',
                obs_key={'choices': config.value_dict["obs"], 'label': "Obs:"},
                operation_type={'choices': ["contains", "is equal to", "is not", "is in"], 'label': 'Operation:'},
                obs_value={'label': 'Value:'},
                reset={'label': 'Reset'}
                )
            def filter_cells_widget(
                obs_key=None,
                operation_type="contains",
                obs_value: str = "",
                reset: bool = False,
                viewer=viewer
            ):
                # find currently selected layer
                layers = viewer.layers
                selected_layers = list(layers.selection)

                if not reset:
                    # create filtering mask
                    if operation_type == "contains":
                        mask = config.adata.obs[obs_key].str.contains(obs_value)
                    elif operation_type == "is equal to":
                        mask = config.adata.obs[obs_key].astype(str) == str(obs_value)
                    elif operation_type == "is not":
                        mask = config.adata.obs[obs_key].astype(str) != str(obs_value)
                    elif operation_type == "is in":
                        obs_value_list = [elem.strip().strip("'").strip('"') for elem in obs_value.split(",")]
                        mask = config.adata.obs[obs_key].isin(obs_value_list)
                    else:
                        raise ValueError(f"Unknown operation type: {operation_type}.")

                    # iterate through selected layers
                    for current_layer in selected_layers:
                        if isinstance(current_layer, napari.layers.points.points.Points):
                            # set visibility
                            fc = current_layer.face_color.copy()
                            fc[:, -1] = 0.
                            fc[mask, -1] = 1.
                            current_layer.face_color = fc
                else:
                    for current_layer in selected_layers:
                        # reset visibility
                        fc = current_layer.face_color.copy()
                        fc[:, -1] = 1.
                        current_layer.face_color = fc

            @add_cells_widget.key.changed.connect
            @add_cells_widget.call_button.changed.connect
            def update_values_on_key_change(event=None):
                _update_values_on_key_change(add_cells_widget)

            @magicgui(
                call_button='Show',
                cell={'label': "Cell:"},
                zoom={'label': 'Zoom:'},
                highlight={'label': 'Highlight'}
                )
            def move_to_cell_widget(
                cell="",
                zoom=5,
                highlight=True,
                ) -> Optional[napari.types.LayerDataTuple]:
                if cell in config.adata.obs_names.astype(str):
                    # get location of selected cell
                    cell_loc = config.adata.obs_names.get_loc(cell)
                    cell_position = config.points[cell_loc]

                    # move center of camera to cell position
                    viewer.camera.center = (0, cell_position[0], cell_position[1])
                    viewer.camera.zoom = zoom

                    if highlight:
                        name = f"cell-{cell}"
                        if name not in viewer.layers:
                            viewer.add_points(
                                data=np.array([cell_position]),
                                name=name,
                                size=6,
                                face_color=[0,0,0,0],
                                opacity=1,
                                edge_color='red',
                                edge_width=0.1
                            )
                else:
                    print(f"Cell '{cell}' not found in `.cells.matrix.obs_names()`.")

            def callback_refresh(event=None):
                # after the points widget is run, the widgets have to be refreshed to current data layer
                _refresh_widgets_after_data_change(xdata,
                                                        points_widget=add_cells_widget,
                                                        boundaries_widget=add_boundaries_widget,
                                                        filter_widget=filter_cells_widget
                                                        )

            def callback_update_legend(event=None):
                _update_colorlegend()

            if add_cells_widget is not None:
                add_cells_widget.call_button.clicked.connect(callback_refresh)
                add_cells_widget.call_button.clicked.connect(callback_update_legend)
            if add_boundaries_widget is not None:
                add_boundaries_widget.call_button.clicked.connect(callback_refresh)
                add_boundaries_widget.call_button.clicked.connect(callback_update_legend)

            viewer.layers.selection.events.active.connect(callback_update_legend)

        if xdata.annotations is None and xdata.regions is None:
            show_geometries_widget = None
        else:
            #TODO: The following section is weirdly complicated and should be simplified.
            # check which geometries are available
            if xdata.annotations is not None:
                if xdata.regions is not None:
                    choices = ["Annotations", "Regions"]
                else:
                    choices = ["Annotations"]
            else:
                choices = ["Regions"]

            for c in choices:
                if len(getattr(xdata, c.lower()).metadata.keys()) == 0:
                    choices.remove(c)

            if len(choices) == 0:
                show_geometries_widget = None
            else:

                # extract geometry object
                geom = getattr(xdata, choices[0].lower())

                # extract annotations keys
                annot_keys = list(geom.metadata.keys())
                try:
                    first_annot_key = list(annot_keys)[0] # for dropdown menu
                except IndexError:
                    show_geometries_widget = None
                else:
                    first_classes = ["all"] + sorted(geom.metadata[first_annot_key]['classes'])

                    @magicgui(
                        call_button='Show',
                        geom_type={"choices": choices, "label": "Type:"},
                        key={"choices": annot_keys, "label": "Key:"},
                        annot_class={"choices": first_classes, "label": "Class:"},
                        edge_width={'min': 1, 'max': 40, 'step': 10, 'label': 'Edge width:'},
                        opacity={'min': 0.0, 'max': 1.0, 'step': 0.1, 'label': 'Opacity:'},
                        show_names={'label': 'Show names'}
                    )
                    def show_geometries_widget(
                        geom_type,
                        key,
                        annot_class,
                        edge_width: int = 10,
                        opacity: float = 1,
                        tolerance: Number = 5,
                        show_names: bool = False
                        ):

                        if geom_type == "Annotations":
                            # get annotation dataframe
                            annot_df = xdata.annotations[key]
                            all_keys = list(xdata.annotations.metadata.keys())
                        elif geom_type == "Regions":
                            # get regions dataframe
                            annot_df = xdata.regions[key]
                            all_keys = list(xdata.regions.metadata.keys())
                        else:
                            TypeError(f"Unknown geometry type: {geom_type}")

                        if annot_class == "all":
                            # get classes
                            classes = annot_df['name'].unique()
                        else:
                            classes = [annot_class]

                        # iterate through classes
                        for cl in classes:

                            layer_name = f"{cl} ({key})"

                            if layer_name not in viewer.layers:
                                # get dataframe for this class
                                class_df = annot_df[annot_df["name"] == cl].copy()

                                # simplify polygons for visualization
                                # class_df["geometry"] = class_df["geometry"].simplify(tolerance)

                                if not "color" in class_df.columns:
                                    # create a RGB color with range 0-255 for this key
                                    rgb_color = [elem * 255 for elem in REGION_CMAP(all_keys.index(key))][:3]
                                else:
                                    rgb_color = None

                                # add layer to viewer
                                _add_geometries_as_layer(
                                    dataframe=class_df,
                                    viewer=viewer,
                                    layer_name=layer_name,
                                    #scale_factor=scale_factor,
                                    edge_width=edge_width,
                                    opacity=opacity,
                                    rgb_color=rgb_color,
                                    show_names=show_names,
                                    mode=geom_type,
                                    tolerance=tolerance
                                )

                    # connect key change with update function
                    @show_geometries_widget.geom_type.changed.connect
                    @show_geometries_widget.key.changed.connect
                    @show_geometries_widget.call_button.clicked.connect
                    @viewer.layers.events.removed.connect # somehow the values change when layers are inserted
                    @viewer.layers.events.inserted.connect # or remoed. Therefore, this update is necessary
                    def update_annotation_widget_after_changes(event=None):
                        _update_keys_based_on_geom_type(show_geometries_widget, xdata=xdata)
                        _update_classes_on_key_change(show_geometries_widget, xdata=xdata)
                        _set_show_names_based_on_geom_type(show_geometries_widget)
                        _update_values_on_key_change(add_cells_widget)

        return add_cells_widget, move_to_cell_widget, show_geometries_widget, add_boundaries_widget, select_data, filter_cells_widget #add_genes, add_observations


    @magic_factory(
        call_button='Add geometry layer',
        key={"choices": ["Geometric annotations", "Point annotations", "Regions"], "label": "Type:"},
        annot_key={'label': 'Key:'},
        class_name={'label': 'Class:'}
        )
    def add_new_geometries_widget(
        key: str = "Geometric annotations",
        annot_key: str = "TestKey",
        class_name: str = "TestClass",
    ) -> napari.types.LayerDataTuple:
        # name pattern of layer name
        name_pattern: str = "{type_symbol} {class_name} ({annot_key})"

        if (class_name != "") & (annot_key != ""):
            if key == "Geometric annotations":
                # generate name
                name = name_pattern.format(
                    type_symbol=ANNOTATIONS_SYMBOL,
                    class_name=class_name,
                    annot_key=annot_key
                    )

                # generate shapes layer for geometric annotation
                layer = (
                    [],
                    {
                        'name': name,
                        'shape_type': 'polygon',
                        'edge_width': 10,
                        'edge_color': 'red',
                        'face_color': 'transparent',
                        #'scale': (config.pixel_size, config.pixel_size),
                        'properties': {
                            'uid': np.array([], dtype='object')
                        }
                        },
                    'shapes'
                    )
            elif key == "Point annotations":
                # generate name
                name = name_pattern.format(
                    type_symbol=POINTS_SYMBOL,
                    class_name=class_name,
                    annot_key=annot_key
                    )

                # generate points layer for point annotation
                layer = (
                    [],
                    {
                        'name': name,
                        'size': 40,
                        'edge_color': 'black',
                        'face_color': 'blue',
                        #'scale': (config.pixel_size, config.pixel_size),
                        'properties': {
                            'uid': np.array([], dtype='object')
                        }
                        },
                    'points'
                    )

            elif key == "Regions":
                # generate name
                name = name_pattern.format(
                    type_symbol=REGIONS_SYMBOL,
                    class_name=class_name,
                    annot_key=annot_key
                    )

                # generate shapes layer for region
                layer = (
                    [],
                    {
                        'name': name,
                        'shape_type': 'polygon',
                        'edge_width': 10,
                        'edge_color': '#ffaa00ff',
                        'face_color': 'transparent',
                        #'scale': (config.pixel_size, config.pixel_size),
                        'properties': {
                            'uid': np.array([], dtype='object')
                        }
                        },
                    'shapes'
                    )

            else:
                layer = None

            # reset class name to nothing
            add_new_geometries_widget.class_name.value = ""

            return layer

        else:
            return None



import os
import warnings
from copy import deepcopy
from numbers import Number
from os.path import relpath
from pathlib import Path
from typing import List, Literal, Optional, Tuple, Union

import cv2
import dask.array as da
import geopandas as gpd
import numpy as np
import pandas as pd
import zarr
from anndata import AnnData
from parse import *
from shapely import MultiPoint, MultiPolygon, Point, Polygon, affinity

from insitupy import __version__
from insitupy._constants import FORBIDDEN_ANNOTATION_NAMES
from insitupy.images.utils import resize_image
from insitupy.utils.utils import convert_int_to_xenium_hex

from .._exceptions import InvalidDataTypeError, InvalidFileTypeError
from ..images.io import read_image, write_ome_tiff, write_zarr
from ..images.utils import create_img_pyramid, crop_dask_array_or_pyramid
from ..io.files import check_overwrite_and_remove_if_true, write_dict_to_json
from ..io.geo import parse_geopandas, write_qupath_geojson
from ..utils.utils import convert_to_list, decode_robust_series
from ..utils.utils import textformat as tf
from ._mixins import DeepCopyMixin


class ShapesData(DeepCopyMixin):
    '''
    Object to store annotations.
    '''
    # default_skip_multipolygons = False
    def __init__(self,
                 files: Optional[List[Union[str, os.PathLike, Path]]] = None,
                 keys: Optional[List[str]] = None,
                 pixel_size: Optional[float] = None,
                 assert_uniqueness: Optional[bool] = None,
                #  skip_multipolygons: Optional[bool] = None,
                 polygons_only: Optional[bool] = None,
                 forbidden_names: Optional[List[str]] = None,
                 shape_name: Optional[str] = None
                 ) -> None:

        self._default_assert_uniqueness = False
        self._default_polygons_only = False
        self._repr_color = tf.Cyan
        self._default_forbidden_names = None

        self._shape_name = shape_name if shape_name is not None else "shapes"

        # create dictionary for metadata
        self._metadata = {}

        self._data =dict()

        # set configuration of ShapesData
        if assert_uniqueness is None:
            self._assert_uniqueness = self._default_assert_uniqueness
        else:
            self._assert_uniqueness = assert_uniqueness

        # if skip_multipolygons is None:
        #     self.skip_multipolygons = self.default_skip_multipolygons

        if polygons_only is None:
            self._polygons_only = self._default_polygons_only
        else:
            self._polygons_only = polygons_only

        if forbidden_names is None:
            self._forbidden_names = self._default_forbidden_names
        else:
            self._forbidden_names = forbidden_names

        if files is not None:
            # make sure files and keys are a list
            assert keys is not None, "If `files` are given, also corresponding `keys` need to be given."
            files = convert_to_list(files)
            keys = convert_to_list(keys)
            assert len(files) == len(keys), "Number of files does not match number of keys."

            assert pixel_size is not None, "If files and `keys` are given, also `pixel_size` needs to be specified"

            if files is not None:
                for key, file in zip(keys, files):
                    # read annotation and store in dictionary
                    self.add_data(data=file,
                                        key=key,
                                        scale_factor=pixel_size,
                                        )

    def __repr__(self):
        if len(self._metadata) > 0:
            repr_strings = []
            for l, m in self._metadata.items():
                # add ' to classes
                classes = [f"'{elem}'" for elem in m["classes"]]
                lc = len(classes)

                # create string
                r = (
                    f'{tf.Bold}{l}:{tf.ResetAll}\t{m[f"n_{self._shape_name}"]} '
                    f'{self._shape_name}, {lc} '
                    f'{"classes" if lc>1 else "class"} '
                )
                if lc < 10:
                    r += f'({",".join(classes)}) {m["analyzed"]}'
                repr_strings.append(r)
            # repr_strings = [
            #     #f'{tf.Bold}{l}:{tf.ResetAll}\t{m[f"n_{self.shape_name}"]} {self.shape_name}, {len(m["classes"])} classes {*m["classes"],} {m["analyzed"]}' for l, m in self.metadata.items()
            #     f'{tf.Bold}{l}:{tf.ResetAll}\t{m[f"n_{self.shape_name}"]} {self.shape_name}, {len(m["classes"])} {"classes" if len(m["classes"])>1 else "class"} ({",".join(m["classes"])}) {m["analyzed"]}' for l, m in self.metadata.items()
            #     ]

            s = "\n".join(repr_strings)
        else:
            s = ""
        repr = f"{self._repr_color}{tf.Bold}{self._shape_name}{tf.ResetAll}\n{s}"
        return repr


    def __getitem__(self, key):
        return self._data.get(key)

    @property
    def metadata(self):
        return self._metadata

    @metadata.setter
    def metadata(self, value: dict):
        self._metadata = value

    def _check_uniqueness(self,
                          dataframe: Optional[gpd.GeoDataFrame] = None,
                          key: Optional[str] = None,
                          verbose: bool = True
                          ) -> bool:

        if dataframe is None:
            annot_df = self[key]
        else:
            annot_df = dataframe

        if len(annot_df.index.unique()) != len(annot_df.name.unique()):
            warnings.warn(
                message=
                (
                    f"The names of the {self._shape_name} for key '{key}' were not unique and thus "
                    f"the key was skipped. In regions only one geometry per class is allowed."
                    )
                )
            return False
        else:
            if verbose:
                print(f"Names of {self._shape_name} for key '{key}' are unique.")
            return True

    def _update_metadata(self,
                         keys: Union[str, Literal["all"]] = "all",
                         analyzed: bool = False,
                         verbose: bool = False
                         ):

        if keys == "all":
            keys = list(self._metadata.keys())

        keys = convert_to_list(keys)
        keys_to_remove = []
        for key in keys:
            if self[key] is None:
                self._metadata.pop(key)
                if verbose:
                    print(f'Removed {key}', flush=True)
            else:
                annot_df = self[key]
                # record metadata information
                self._metadata[key][f"n_{self._shape_name}"] = len(annot_df)  # number of annotations

                try:
                    self._metadata[key]["classes"] = annot_df['name'].unique().tolist()  # annotation classes
                except KeyError:
                    self._metadata[key]["classes"] = ["unnamed"]

                self._metadata[key]["analyzed"] = tf.Tick if analyzed else ""  # whether this annotation has been used in the annotate() function

    def add_data(self,
                    data: Union[gpd.GeoDataFrame, pd.DataFrame, dict,
                                str, os.PathLike, Path],
                    key: str,
                    scale_factor: Number,
                    default_name: str = "name",
                    verbose: bool = False,
                   ):
        # parse geopandas data from dataframe or file
        new_df = parse_geopandas(data)

        if "name" not in new_df.columns:
            new_df["name"] = ["None"] * len(new_df)

        if self._forbidden_names is not None:
            try:
                new_names = new_df["name"].tolist()
            except KeyError:
                pass
            else:
                if np.any([elem in new_names for elem in self._forbidden_names]):
                    raise ValueError(f"One of the forbidden names for annotations ({self._forbidden_names}) has been used in the imported dataset. Please change the respective change to prevent interference with downstream functions.")

        # convert geometries into unit (e.g. µm) values
        new_df["geometry"] = new_df["geometry"].scale(xfact=scale_factor, yfact=scale_factor, origin=(0,0))

        # if "scale" not in new_df.columns:
        #     # add scale factor to data
        #     if scale_factor is None:
        #         warnings.warn("No `scale_factor` added to data.")
        #     new_df["scale"] = [scale_factor] * len(new_df)
        # else:
        #     if verbose:
        #         print("Scale inferred from file.", flush=True)

        # determine the type of layer that needs to be used in napari later
        layer_types = []
        for geom in new_df["geometry"]:
            if isinstance(geom, Point) or isinstance(geom, MultiPoint):
                layer_types.append("Points")
            else:
                layer_types.append("Shapes")
        new_df["layer_type"] = layer_types

        # # convert pixel coordinates to metric units
        # new_df["geometry"] = new_df.geometry.scale(origin=(0,0), xfact=pixel_size, yfact=pixel_size)

        if key not in self._data.keys():
            # if key does not exist yet, the new df is the whole annotation dataframe
            annot_df = new_df

            # collect additional variables for reporting
            new_geometries_added = True # dataframe will be added later
            existing_str = ""
            old_n = 0
            new_n = len(annot_df)
        else:
            # concatenate old and new annoation dataframe
            annot_df = self[key]
            old_n = len(annot_df)
            annot_df = pd.concat([annot_df, new_df], ignore_index=False)

            # remove all duplicated shapes - leaving only the newly added
            annot_df = annot_df[~annot_df.index.duplicated()]
            new_n = len(annot_df)

            # collect additional variables for reporting
            new_geometries_added = new_n > old_n
            existing_str = "existing "

        if new_geometries_added:
            add = True
            if self._assert_uniqueness:
                # check if the shapes data for this key is unique (same number of names than indices)
                is_unique = self._check_uniqueness(dataframe=annot_df, key=key, verbose=False)

                if not is_unique:
                    add = False

            if self._polygons_only:
                # check if any of the shapes are shapely MultiPolygons
                is_not_polygon = [not isinstance(p, Polygon) for p in annot_df.geometry]
                if np.any(is_not_polygon):
                    annot_df = annot_df.loc[is_not_polygon]
                    warnings.warn(
                        f"Some {self._shape_name} were not pure Polygon objects and skipped.",
                        stacklevel=2
                        )

            # check that the dataframe is not empty
            if len(annot_df) == 0:
                add = False

            if add:
                # add dataframe to AnnotationData object
                self._data[key] = annot_df

                # add new entry to metadata
                self._metadata[key] = {}

                # update metadata
                self._update_metadata(keys=key, analyzed=False)

                if verbose:
                    # report
                    print(f"Added {new_n - old_n} new {self._shape_name} to {existing_str}key '{key}'")

    def crop(self,
             shape,
             xlim,
             ylim,
             verbose=True
             ):
        if shape is None:
            if (xlim is None) or (ylim is None):
                raise ValueError("If shape is None, both xlim and ylim must not be None.")
            else:
                shape = Polygon([(xlim[0], ylim[0]), (xlim[1], ylim[0]), (xlim[1], ylim[1]), (xlim[0], ylim[1])])
        else:
            if (xlim is not None) and (ylim is not None):
                if verbose:
                    warnings.warn("Both xlim/ylim and shape are provided. Shape will be used for cropping.")

        new_metadata = {}
        for i, n in enumerate(self._metadata.keys()):
            shapesdf = self[n]

            # select annotations that intersect with the selected area
            mask = [shape.intersects(elem) for elem in shapesdf["geometry"]]
            shapesdf = shapesdf.loc[mask, :].copy()

            # move origin to zero after cropping
            shapesdf["geometry"] = shapesdf["geometry"].apply(affinity.translate, xoff=-xlim[0], yoff=-ylim[0])

            # check if there are annotations left or if it has to be deleted
            if len(shapesdf) > 0:
                # add new dataframe back to annotations object
                self._data[n] = shapesdf

                # update metadata
                new_metadata[n] = {}
                new_metadata[n][f"n_{self._shape_name}"] = len(shapesdf)
                new_metadata[n]["classes"] = shapesdf.name.unique().tolist()
                new_metadata[n]["analyzed"] = self._metadata[n]["analyzed"]  # analyzed information is just copied

            else:
                # delete annotations
                del self._data[n]

        self._metadata = new_metadata

        self._update_metadata()

    def remove_data(self,
                   key_to_remove: str,
                   classes_to_remove: Union[Literal["all"], List[str], str] = "all"
                   ):
        if classes_to_remove == "all":
            del self._data[key_to_remove]
        else:
            classes_to_remove = convert_to_list(classes_to_remove)
            geom_df = self[key_to_remove]
            self._data[key_to_remove] = geom_df[~geom_df.name.isin(classes_to_remove)]

        self._update_metadata()

    def save(self,
             path: Union[str, os.PathLike, Path],
             overwrite: bool = False
             ):
        path = Path(path)

        # check if the output file should be overwritten
        check_overwrite_and_remove_if_true(path, overwrite=overwrite)

        # create directory
        path.mkdir(parents=True, exist_ok=True)

        # # create path for matrix
        # annot_path = (path / self.shape_name)
        # annot_path.mkdir(parents=True, exist_ok=True) # create directory

        # if metadata is not None:
        #     metadata["annotations"] = {}
        for n in self._metadata.keys():
            df = self[n]
            # annot_file = annot_path / f"{n}.parquet"
            # annot_df.to_parquet(annot_file)
            shapes_file = path / f"{n}.geojson"
            write_qupath_geojson(dataframe=df, file=shapes_file)

            # if metadata is not None:
            #     metadata["annotations"][n] = Path(relpath(annot_file, path)).as_posix()

        # save AnnotationData metadata
        shape_meta_path = path / f"metadata.json"
        write_dict_to_json(dictionary=self._metadata, file=shape_meta_path)

class AnnotationsData(ShapesData):
    def __init__(self,
                 files: Optional[List[Union[str, os.PathLike, Path]]] = None,
                 keys: Optional[List[str]] = None,
                 pixel_size: Optional[float] = None
                 ) -> None:
        self._default_assert_uniqueness = False
        # self.default_skip_multipolygons = False
        self._default_polygons_only = False
        self._shape_name = "annotations"
        self._repr_color = tf.Cyan
        self._default_forbidden_names = FORBIDDEN_ANNOTATION_NAMES

        ShapesData.__init__(self, files, keys, pixel_size, shape_name=self._shape_name)

class RegionsData(ShapesData):
    def __init__(self,
                 files: Optional[List[Union[str, os.PathLike, Path]]] = None,
                 keys: Optional[List[str]] = None,
                 pixel_size: Optional[float] = None
                 ) -> None:
        self._default_assert_uniqueness = True
        # self.default_skip_multipolygons = True # MultiPolygons are not allowed in regions
        self._default_polygons_only = True
        self._shape_name = "regions"
        self._repr_color = tf.Yellow

        ShapesData.__init__(self, files, keys, pixel_size, shape_name=self._shape_name)

class BoundariesData(DeepCopyMixin):
    '''
    Object to read and load boundaries of cells and nuclei.
    '''
    def __init__(self,
                 cell_names: Union[da.core.Array, np.ndarray, List],
                 seg_mask_value: Optional[Union[da.core.Array, np.ndarray, List]],
                 ):
        """
        Initialize the BoundariesData object.

        Args:
            cell_names (Union[da.core.Array, np.ndarray, List]): Cell names which need to correspond to `.obs_names` in the `.matrix` of `CellData`.
            seg_mask_value (Optional[Union[da.core.Array, np.ndarray, List]]): Segmentation mask values. Required to have the same length as `cell_names`.
                Specifies which values in the "cells" segmentation mask correspond to which cell name.

        For more details on how these values are saved in case of Xenium In Situ, see:
        https://www.10xgenomics.com/support/software/xenium-onboard-analysis/latest/tutorials/outputs/xoa-output-zarr
        """
        self._metadata = {}

        # store cell ids
        #self._cell_ids = da.from_array(np.array(cell_ids, dtype=np.uint32))
        self._cell_names = da.from_array(np.array(cell_names, dtype=str))

        self._seg_mask_value = seg_mask_value
        if self._seg_mask_value is not None:
            self._seg_mask_value = da.from_array(np.array(seg_mask_value, dtype=np.uint32))
        else:
            raise ValueError("Argument 'seg_mask_value' is None. This argument is required to be set.")

        self._data = dict()

    def __repr__(self):
        labels = list(self._metadata.keys())
        if len(labels) == 0:
            repr = f"Empty BoundariesData object"
        else:
            ll = len(labels)
            repr = f"BoundariesData object with {ll} {'entry' if ll == 1 else 'entries'}:"
            for l in labels:
                repr += f"\n{tf.SPACER+tf.Bold+l+tf.ResetAll}"
        return repr

    def __getitem__(self, key):
        return self._data.get(key)

    def __setitem__(self, key: str, item):
        self._data[key] = item

    @property
    def metadata(self):
        return self._metadata

    # @property
    # def cell_ids(self):
    #     return self._cell_ids

    @property
    def cell_names(self):
        return self._cell_names

    @property
    def seg_mask_value(self):
        return self._seg_mask_value

    def add_boundaries(self,
                       data: Optional[Union[dict, List[str]]],
                       pixel_size: Number, # required for boundaries that are saved as masks
                       labels: Optional[List[str]] = [],
                       overwrite: bool = False
                       ):
        if isinstance(data, dict):
            # extract keys from dictionary
            labels = data.keys()
            data = data.values()
        elif isinstance(data, list):
            if labels is None:
                raise ValueError("Argument 'labels' is None. If 'dataframes' is a list, 'labels' is required to be a list, too.")
            else:
                # make sure labels is a list
                labels = convert_to_list(labels)
        else:
            data = convert_to_list(data)
            labels = convert_to_list(labels)
            #raise ValueError(f"Argument 'dataframes' has unknown file type ({type(data)}). Expected to be a list or dictionary.")

        for l, df in zip(labels, data):
            if isinstance(df, pd.DataFrame) or isinstance(df, da.core.Array) or np.all([isinstance(elem, da.core.Array) for elem in df]):
                if l not in self._metadata or overwrite:
                    # add to object
                    self._data[l] = df
                    self._metadata[l] = {}
                    self._metadata[l]["pixel_size"] = pixel_size
                else:
                    raise KeyError(f"Label '{l}' exists already in BoundariesData object. To overwrite, set 'overwrite' argument to True.")
            else:
                print(f"Boundaries element `{l}` is neither a pandas DataFrame nor a Dask Array. Was not added.")

    def crop(self,
             cell_ids: List[str],
             xlim: Tuple[int, int],
             ylim: Tuple[int, int]
             ):
        '''
        Crop the BoundariesData object.
        '''
        # make sure cell ids are a list
        cell_ids = convert_to_list(cell_ids)

        for n, meta in self._metadata.items():
            # get dataframe
            data = self[n]

            try:
                # get pixel size
                pixel_size = meta["pixel_size"]

                data = crop_dask_array_or_pyramid(
                    data=data,
                    xlim=xlim,
                    ylim=ylim,
                    pixel_size=pixel_size
                )
            except InvalidDataTypeError:
                # filter dataframe
                data = data.loc[data["cell_id"].isin(cell_ids), :]

                # re-center to 0
                data["vertex_x"] -= xlim[0]
                data["vertex_y"] -= ylim[0]

            # add to object
            self._data[n] = data

    def convert_to_shapely_objects(self):
        for n in self._metadata.keys():
            print(f"Converting `{n}` to GeoPandas DataFrame with shapely objects.")
            # retrief dataframe with boundary coordinates
            df = self[n]

            if isinstance(df, pd.DataFrame):
                # convert xy coordinates into shapely Point objects
                df["geometry"] = gpd.points_from_xy(df["vertex_x"], df["vertex_y"])
                del df["vertex_x"], df["vertex_y"]

                # convert points into polygon objects per cell id
                df = df.groupby("cell_id")['geometry'].apply(lambda x: Polygon(x.tolist()))
                df.index = decode_robust_series(df.index)  # convert byte strings in index

                # add to object
                self._data[n] = pd.DataFrame(df)
            else:
                print(f"Boundaries element `{n} was no Dataframe. Skipped.")

    def save(self,
             bound_file : Union[str, os.PathLike, Path] = "boundaries.zarr.zip",
             save_as_pyramid: bool = True
             ):
        bound_file = Path(bound_file)
        suffix = bound_file.name.split(".", maxsplit=1)[-1]

        if suffix not in ["zarr", "zarr.zip"]:
            raise InvalidFileTypeError(allowed_types=[".zarr", ".zarr.zip"], received_type=suffix)

        with zarr.ZipStore(bound_file, mode='w') if suffix == "zarr.zip" else zarr.DirectoryStore(bound_file) as dirstore:
            # for conditional 'with' see also: https://stackoverflow.com/questions/27803059/conditional-with-statement-in-python
            for n in self._metadata.keys():
                bound_data = self[n]

                # check data
                if isinstance(bound_data, list):
                    if not save_as_pyramid:
                        bound_data = bound_data[0]
                else:
                    if save_as_pyramid:
                        # create pyramid
                        bound_data = create_img_pyramid(img=bound_data, nsubres=6)


                #if isinstance(bound_data, dask.array.core.Array):
                if isinstance(bound_data, list):
                    for i, b in enumerate(bound_data):
                        comp = f"masks/{n}/{i}"
                        b.to_zarr(dirstore, component=comp)
                else:
                    bound_data.to_zarr(dirstore, component=f"masks/{n}")

                # add boundaries metadata to zarr.zip
                store = zarr.open(dirstore, mode="a")
                store[f"masks/{n}"].attrs.put(self._metadata[n])

                # save keys in insitupy metadata
                #metadata["boundaries"]["keys"].append(n)

            # save paths in insitupy metadata
            #metadata["boundaries"]["path"] = Path(relpath(bound_file, path)).as_posix()

            #self._cell_ids.to_zarr(dirstore, component="cell_id")
            self.cell_names.to_zarr(dirstore, component="cell_names", overwrite=True)

            if self._seg_mask_value is not None:
                self.seg_mask_value.to_zarr(dirstore, component="seg_mask_value", overwrite=True)

        # # add version to metadata
        # metadata_to_save = self.metadata.copy()
        # metadata_to_save["version"] = __version__

        # # save metadata
        # write_dict_to_json(dictionary=metadata_to_save, file=path / ".boundariesdata")

class CellData(DeepCopyMixin):
    '''
    Data object containing an AnnData object and a boundary object which are kept in sync.
    '''
    def __init__(self,
               matrix: AnnData,
               boundaries: Optional[BoundariesData],
               config: dict = {}
               ):
        self._matrix = matrix
        self._config = config

        if boundaries is not None:
            self._boundaries = boundaries
            self._entries = ["matrix", "boundaries"]
        else:
            self._boundaries = None
            self._entries = ["matrix"]

    def __repr__(self):
        repr = (
            f"{tf.Bold+'matrix'+tf.ResetAll}\n"
            f"{tf.SPACER+self._matrix.__repr__()}"
        )

        if self._boundaries is not None:
            bound_repr = self._boundaries.__repr__()

            repr += f"\n{tf.Bold+'boundaries'+tf.ResetAll}\n" + tf.SPACER + bound_repr.replace("\n", f"\n{tf.SPACER}")
        return repr

    @property
    def matrix(self):
        return self._matrix

    @matrix.setter
    def matrix(self, value: AnnData):
        self._matrix = value

    @property
    def config(self):
        return self._config

    @property
    def boundaries(self):
        return self._boundaries

    @property
    def entries(self):
        return self._entries

    def copy(self):
        '''
        Function to generate a deep copy of the current object.
        '''

        return deepcopy(self)

    def crop(self,
            xlim: Optional[Tuple[int, int]] = None,
            ylim: Optional[Tuple[int, int]] = None,
            shape: Optional[Union[Polygon, MultiPolygon]] = None,
            inplace: bool = False,
            verbose: bool = True
            ):

        # check if the changes are supposed to be made in place or not
        if inplace:
            _self = self
        else:
            _self = self.copy()

        # retrieve cell coordinates
        cell_coords = _self.matrix.obsm['spatial'].copy()

        # Ensure that either both xlim and ylim are not None or shape is not None
        if (xlim is None or ylim is None) and shape is None:
            raise ValueError("Either both xlim and ylim must be provided, or shape must be provided.")
        # if xlim is not None and ylim is not None and shape is not None:
        #     warnings.warn("Both xlim/ylim and shape are provided. Shape will be used for cropping.")

        if shape is not None:
            if xlim is not None and ylim is not None:
                if verbose:
                    warnings.warn("Both xlim/ylim and shape are provided. Shape will be used for cropping.")

            # create shapely objects from cell coordinates
            cells = gpd.points_from_xy(cell_coords[:, 0], cell_coords[:, 1])

            # create a mask based on the shape
            mask = shape.contains(cells)

            # get bounding box of shape
            minx, miny, maxx, maxy = shape.bounds # (minx, miny, maxx, maxy)
            xlim = (minx, maxx)
            ylim = (miny, maxy)

        else:
            if xlim is None or ylim is None:
                raise ValueError("Either both xlim and ylim must be provided, or shape must be provided.")

            # make sure there are no negative values in the limits
            xlim = tuple(np.clip(xlim, a_min=0, a_max=None))
            ylim = tuple(np.clip(ylim, a_min=0, a_max=None))

            # create a mask based on xlim and ylim
            xmask = (cell_coords[:, 0] >= xlim[0]) & (cell_coords[:, 0] <= xlim[1])
            ymask = (cell_coords[:, 1] >= ylim[0]) & (cell_coords[:, 1] <= ylim[1])
            mask = xmask & ymask

        # select
        _self.matrix = _self.matrix[mask, :].copy()

        # crop boundaries
        _self.boundaries.crop(
            cell_ids=_self.matrix.obs_names, xlim=xlim, ylim=ylim
            )

        # shift coordinates to correct for change of coordinates during cropping
        if shape is not None:
            minx, miny, _, _ = shape.bounds
            _self.shift(x=-minx, y=-miny)
        else:
            _self.shift(x=-xlim[0], y=-ylim[0])

        # sync the ids and names
        _self.sync_cell_ids()

    def save(self,
             path: Union[str, os.PathLike, Path],
             boundaries_zipped: bool = False,
             boundaries_as_pyramid: bool = True,
             overwrite: bool = False,
             ):

        path = Path(path)
        celldata_metadata = {}

        # check if the output file should be overwritten
        check_overwrite_and_remove_if_true(path, overwrite=overwrite)

        # create directory
        path.mkdir(parents=True, exist_ok=True)

        # # create path for matrix
        # mtx_path = path / "matrix"
        # mtx_path.mkdir(parents=True, exist_ok=True) # create directory

        # write matrix to file
        mtx_file = path / "matrix.h5ad"
        self._matrix.write(mtx_file)
        celldata_metadata["matrix"] = Path(relpath(mtx_file, path)).as_posix()

        # save boundaries
        if self._boundaries is not None:
            boundaries = self._boundaries
            if boundaries_zipped:
                bound_file = path / "boundaries.zarr.zip"
            else:
                bound_file = path / "boundaries.zarr"

            # save boundaries
            boundaries.save(bound_file, save_as_pyramid=True)

            # add entry for boundaries to metadata
            celldata_metadata["boundaries"] = Path(relpath(bound_file, path)).as_posix()
            # bound_path.mkdir(parents=True, exist_ok=True) # create directory

        # add version to metadata
        celldata_metadata["version"] = __version__

        # add configurations
        if self._config is not None:
            celldata_metadata["config"] = self._config

        # save metadata
        write_dict_to_json(dictionary=celldata_metadata, file=path / ".celldata")


    def sync_cell_ids(self):
        '''
        Function to synchronize matrix and boundaries of CellData.

        Procedure:
        1. Select matrix cell IDs
        2. Check if all matrix cell IDs are in boundaries
            - if not all are in boundaries, throw error saying that those will also be removed
        3. Select only matrix cell IDs which are also in boundaries and filter for them
        '''
        # get cell IDs from matrix
        matrix_cell_ids_hex = self._matrix.obs_names.astype(str)

        if self._boundaries is None:
            print('No `boundaries` attribute found in CellData found.')
        else:
            boundaries = self._boundaries

            # create pandas series from seg_mask values and cell_names
            ds = pd.Series(
                data=boundaries.seg_mask_value,
                index=boundaries.cell_names
                )

            filter_mask_in = ds.index.isin(matrix_cell_ids_hex)

            # filter cell names and seg_mask_values
            boundaries._seg_mask_value = da.from_array(np.array(ds[filter_mask_in]))
            boundaries._cell_names = da.from_array(np.array(ds.index[filter_mask_in], dtype=str))

            # find the seg_mask_values which are not anymore present
            seg_mask_values_not_in_matrix = ds[~filter_mask_in].values

            cell_bounds = boundaries["cells"]
            nuc_bounds = boundaries["nuclei"]

            if isinstance(cell_bounds, list):
                assert isinstance (nuc_bounds, list), "Cellular boundaries are a image pyramid but nuclear boundaries are not. Both need to be of the same type for the synchronization to work."
                for cell_bound, nuc_bound in zip(cell_bounds, nuc_bounds):
                    removed_cells_mask = da.isin(cell_bound, seg_mask_values_not_in_matrix)
                    cell_bound[removed_cells_mask] = 0 # set all removed cells 0
                    nuc_bound[removed_cells_mask] = 0 # set all nuclei belong to the removed cells 0
            elif isinstance(cell_bounds, da.core.Array):
                assert isinstance (nuc_bounds, da.core.Array), "Cellular boundaries are a dask array but nuclear boundaries are not. Both need to be of the same type for the synchronization to work."
                # set all non existent cell ids to zero
                removed_cells_mask = da.isin(cell_bounds, seg_mask_values_not_in_matrix)
                cell_bounds[removed_cells_mask] = 0 # set all removed cells 0
                nuc_bounds[removed_cells_mask] = 0 # set all nuclei belong to the removed cells 0
            else:
                warnings.warn(f"Unknown data type for cellular boundaries: {type(cell_bounds)}. Need to be either a dask array or a list of dask arrays. Skipped synchronization of cell ids.")

    def shift(self,
              x: Union[int, float],
              y: Union[int, float]
              ):
        '''
        Function to shift the coordinates of both matrix and boundaries data by certain values x/y.
        '''

        # move origin again to 0 by subtracting the lower limits from the coordinates
        cell_coords = self._matrix.obsm['spatial'].copy()
        cell_coords[:, 0] += x
        cell_coords[:, 1] += y
        self._matrix.obsm['spatial'] = cell_coords

        if self._boundaries is None:
            print('No `boundaries` attribute found in CellData found.')
        else:
            boundaries = self._boundaries
            for n in boundaries.metadata.keys():
                # get dataframe
                df = boundaries[n]

                if isinstance(df, pd.DataFrame):
                    # re-center to 0
                    df["vertex_x"] += x
                    df["vertex_y"] += y

                    # add to object
                    setattr(self._boundaries, n, df)


class ImageData(DeepCopyMixin):
    '''
    Object to read and load images.
    '''
    def __init__(self,
                 #path: Union[str, os.PathLike, Path] = None,
                 img_files: List[str] = None,
                 img_names: List[str] = None,
                 pixel_size: float = None,
                 ):
        # # add path to object
        # self.path = path

        # iterate through files and load them
        self._names = []
        self._metadata = {}
        self._data = dict()

        if img_files is not None:
            # convert arguments to lists
            img_files = convert_to_list(img_files)
            img_names = convert_to_list(img_names)

            for n, f in zip(img_names, img_files):
                #impath = path / f
                self.add_image(
                    image=f,
                    name=n,
                    axes=None,
                    pixel_size=pixel_size,
                    ome_meta=None
                    )

    def __repr__(self):
        if len(self._metadata) > 0:
            repr_strings = [f"{tf.Bold}{n}:{tf.ResetAll}\t{metadata['shape']}" for n,metadata in self._metadata.items()]
            s = "\n".join(repr_strings)
        else:
            s = "empty"
        repr = f"{tf.Blue+tf.Bold}images{tf.ResetAll}\n{s}"
        return repr

    def __getitem__(self, key):
        return self._data.get(key)

    @property
    def metadata(self):
        return self._metadata

    @property
    def names(self):
        return self._names

    def add_image(
        self,
        image: Union[da.core.Array, np.ndarray, str, os.PathLike, Path],
        name: str,
        axes: Optional[str] = None, # channels - other examples: 'TCYXS'. S for RGB channels. 'YX' for grayscale image.
        pixel_size: Optional[Number] = None,
        ome_meta: Optional[dict] = None,
        overwrite: bool = False,
        verbose: bool = True
        ):
        if name in self._names:
            if not overwrite:
                print(f"`ImageData` object contains already an image with name '{name}'. Image is not added.") if verbose else None
                do_addition = False
            else:
                # remove attribute with current name
                del self._data[name]

                # remove from name list and metadata
                self._names = [elem for elem in self._names if elem != name]
                self._metadata.pop(name, None)

                do_addition = True
        else:
            do_addition = True

        if do_addition:
            # check if image is a path or a data array
            if isinstance(image, da.core.Array) or isinstance(image, np.ndarray):
                assert axes is not None, "If `image` is numpy or dask array, `axes` needs to be set."
                assert pixel_size is not None, "If `image` is numpy or dask array, `pixel_size` needs to be set."

                try:
                    # convert to dask array before addition
                    img = da.from_array(image)
                except ValueError:
                    # in this case the array was already a dask array
                    img = image
                filename = None

            elif Path(str(image)).exists():
                # read path
                image = Path(image)
                image = image.resolve() # resolve relative path
                filename = image.name
                img, ome_meta, axes = read_image(image)

            else:
                raise ValueError(f"`image` is neither a dask array nor an existing path.")

            # set attribute and add names to object
            self._data[name] = img
            self._names.append(name)

            # retrieve metadata
            img_shape = img[0].shape if isinstance(img, list) else img.shape
            # img_max = img[0].max() if isinstance(img, list) else img.max()
            # try:
            #     img_max = img_max.compute()
            # except AttributeError:
            #     img_max = img_max

            # save metadata
            self._metadata[name] = {}
            self._metadata[name]["filename"] = filename
            self._metadata[name]["shape"] = img_shape  # store shape
            self._metadata[name]["axes"] = axes
            self._metadata[name]["OME"] = ome_meta

            # add universal pixel size to metadata
            try:
                self._metadata[name]['pixel_size'] = float(ome_meta['Image']['Pixels']['PhysicalSizeX'])
            except KeyError:
                self._metadata[name]['pixel_size'] = float(ome_meta['PhysicalSizeX'])

            # check whether the image is RGB or not
            if len(img_shape) == 3:
                self._metadata[name]["rgb"] = True
            elif len(img_shape) == 2:
                self._metadata[name]["rgb"] = False
            else:
                raise ValueError(f"Unknown image shape: {img_shape}")

            # # get image contrast limits
            # if self._metadata[name]["rgb"]:
            #     self._metadata[name]["contrast_limits"] = (0, img_max)
            # else:
            #     self._metadata[name]["contrast_limits"] = (0, img_max)


    def load(self,
             which: Union[List[str], str] = "all"
             ):
        '''
        Load images into memory.
        '''
        if which == "all":
            which = self._img_names

        # make sure which is a list
        which = convert_to_list(which)
        for n in which:
            img_loaded = self[n].compute()
            self._data[n] = img_loaded

    def crop(self,
             xlim: Optional[Tuple[int, int]],
             ylim: Optional[Tuple[int, int]]
             ):
        # extract names from metadata
        names = list(self._metadata.keys())
        for n in names:
            # extract the image pyramid
            img_data = self[n]

            # extract pixel size
            pixel_size = self._metadata[n]['pixel_size']

            cropped_img_data = crop_dask_array_or_pyramid(
                data=img_data,
                xlim=xlim,
                ylim=ylim,
                pixel_size=pixel_size
            )

            # save cropping properties in metadata
            self._metadata[n]["cropping_xlim"] = xlim
            self._metadata[n]["cropping_ylim"] = ylim

            try:
                self._metadata[n]["shape"] = cropped_img_data.shape
            except AttributeError:
                self._metadata[n]["shape"] = cropped_img_data[0].shape

            # add cropped pyramid to object
            self._data[n] = cropped_img_data

    def save(self,
             output_folder: Union[str, os.PathLike, Path],
             keys_to_save: Optional[str] = None,
             as_zarr: bool = True,
             zipped: bool = False,
             save_pyramid: bool = True,
             compression: Literal['jpeg', 'LZW', 'jpeg2000', 'ZLIB', None] = 'ZLIB', # jpeg2000 or ZLIB are recommended in the Xenium documentation - ZLIB is faster
             return_savepaths: bool = False,
             overwrite: bool = False,
             max_resolution: Optional[Number] = None, # in µm per pixel
             verbose: bool = False
             ):
        """
        Save images to the specified output folder in either Zarr or OME-TIFF format.

        Args:
            output_folder (Union[str, os.PathLike, Path]): The directory where images will be saved.
            keys_to_save (Optional[str]): Specific keys of images to save. If None, all images are saved.
            as_zarr (bool): If True, save images in Zarr format. Otherwise, save as OME-TIFF.
            zipped (bool): If True and saving as Zarr, compress the Zarr files into zip archives.
            save_pyramid (bool): If True, save image pyramids (only applicable for Zarr format).
            compression (Literal['jpeg', 'LZW', 'jpeg2000', 'ZLIB', None]): Compression method for OME-TIFF files.
            return_savepaths (bool): If True, return the paths of the saved files.
            overwrite (bool): If True, overwrite existing files in the output folder. Default is False.
            max_resolution (Optional[Number]): Maximum resolution for images in µm per pixel. If the pixel size of an image is larger than `max_resolution`, the image is downscaled. Default is None.
            verbose (bool): If True, print status messages during saving. Default is True.

        Returns:
            Optional[Dict[str, Path]]: A dictionary mapping image keys to their save paths if `return_savepaths` is True. Otherwise, returns None.

        Raises:
            FileExistsError: If `overwrite` is False and the output folder already contains files.

        """
        output_folder = Path(output_folder)

        if keys_to_save is None:
            keys_to_save = list(self._metadata.keys())
        else:
            keys_to_save = convert_to_list(keys_to_save)

        # check overwrite
        check_overwrite_and_remove_if_true(path=output_folder, overwrite=overwrite)

        # create output directory
        output_folder.mkdir(parents=True, exist_ok=True)

        if return_savepaths:
            savepaths = {}

        for name, img_metadata in self._metadata.items():
            if name in keys_to_save:
                # extract image
                img = self[name]
                new_img_metadata = img_metadata.copy()

                axes = new_img_metadata['axes']
                pixel_size = new_img_metadata['pixel_size'] # in µm per pixel

                if max_resolution is not None:
                    if max_resolution == pixel_size:
                        warnings.warn(f"`max_pixel_size` ({max_resolution}) equal to `pixel_size` ({pixel_size}). Skipped resizing.")
                        pass
                    if max_resolution < pixel_size:
                        warnings.warn(f"`max_pixel_size` ({max_resolution}) smaller than `pixel_size` ({pixel_size}). Skipped resizing.")
                        pass
                    else:
                        # downscale image
                        if isinstance(img, list):
                            img = img[0]
                        downscale_factor = max_resolution / pixel_size

                        if verbose:
                            print(f"Downscale image to {max_resolution} µm per pixel by factor {downscale_factor}")
                        img = resize_image(img, scale_factor=1/downscale_factor, axes=axes)
                        img = da.from_array(img)

                        # change metadata
                        new_img_metadata['pixel_size'] = max_resolution
                        try:
                            new_img_metadata['OME']['Image']['Pixels']['PhysicalSizeX'] = str(max_resolution)
                        except KeyError:
                            new_img_metadata['OME']['PhysicalSizeX'] = str(max_resolution)

                        try:
                            new_img_metadata['OME']['Image']['Pixels']['PhysicalSizeY'] = str(max_resolution)
                        except KeyError:
                            new_img_metadata['OME']['PhysicalSizeY'] = str(max_resolution)

                if as_zarr:
                    # generate filename
                    if zipped:
                        #filename = Path(img_metadata["file"]).name.split(".")[0] + ".zarr.zip"
                        filename = name + ".zarr.zip"
                    else:
                        # filename = Path(img_metadata["file"]).name.split(".")[0] + ".zarr"
                        filename = name + ".zarr"

                    # write to zarr
                    img_path = output_folder / filename
                    write_zarr(image=img, file=img_path,
                               img_metadata=new_img_metadata,
                               save_pyramid=save_pyramid,
                               axes=axes, verbose=verbose
                               )
                else:
                    # get file name for saving
                    #filename = Path(img_metadata["file"]).name.split(".")[0] + ".ome.tif"
                    filename = name + ".ome.tif"
                    # retrieve image metadata for saving
                    photometric = 'rgb' if new_img_metadata['rgb'] else 'minisblack'


                    # retrieve OME metadata
                    ome_meta_to_retrieve = ["SignificantBits", "PhysicalSizeX", "PhysicalSizeY",
                                            "PhysicalSizeXUnit", "PhysicalSizeYUnit"]

                    try:
                        pixel_meta = new_img_metadata["OME"]["Image"]["Pixels"]
                    except KeyError:
                        pixel_meta = new_img_metadata["OME"]

                    selected_metadata = {key: pixel_meta[key] for key in ome_meta_to_retrieve if key in pixel_meta}

                    # write images as OME-TIFF
                    write_ome_tiff(image=img, file=output_folder / filename,
                                photometric=photometric, axes=axes,
                                compression=compression,
                                metadata=selected_metadata, overwrite=False,
                                verbose=verbose
                                )

                if return_savepaths:
                    # collect savepaths
                    savepaths[name] = output_folder / filename

        if return_savepaths:
            return savepaths
